/* ////////////////////////////////////////////////////////////

  File Name: config.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "config.h"

#include <ysport.h>

PolyCreConfig::PolyCreConfig()
{
	Initialize();
}

PolyCreConfig::~PolyCreConfig()
{
}

/*static*/ PolyCreConfig *PolyCreConfig::Create(void)
{
	return new PolyCreConfig;
}

/*statid*/ void PolyCreConfig::Delete(PolyCreConfig *ptr)
{
	delete ptr;
}

void PolyCreConfig::Initialize(void)
{
	SetDefault();
}

void PolyCreConfig::SetDefault(void)
{
	coordSys=YSOPENGL;
	YsSpecialPath::GetUserDocDir(defDataDir);
	perspective=YSFALSE;

	bgColor[0].SetDoubleRGBA(0.80,0.90,1.00,1.0);
	bgColor[1]=YsWhite();
	bgColor[2].SetDoubleRGBA(0.75,0.75,0.75,1.0);
}

YSRESULT PolyCreConfig::Load(YsTextInputStream &inStream)
{
	YsString str;
	YsArray <YsString,16> args;
	while(NULL!=inStream.Gets(str))
	{
		if(YSOK==str.Arguments(args) && 0<args.GetN())
		{
			if(0==strcmp(args[0],"COORDSYS") && 2<=args.GetN())
			{
				if(0==strcmp(args[1],"LEFT"))
				{
					coordSys=YSLEFT_ZPLUS_YPLUS;
				}
				else if(0==strcmp(args[1],"RIGHT"))
				{
					coordSys=YSRIGHT_ZMINUS_YPLUS;
				}
			}
			else if(0==strcmp(args[0],"PERSPECTIVE") && 2<=args.GetN())
			{
				perspective=YsStrToBool(args[1]);
			}
			else if(0==strcmp(args[0],"DATADIR") && 2<=args.GetN())
			{
				defDataDir.SetUTF8String(args[1]);
			}
			else if(0==strcmp(args[0],"BGCOLOR") && 5<=args.GetN())
			{
				const int colorIdx=atoi(args[1]);
				if(0<=colorIdx && colorIdx<3)
				{
					bgColor[colorIdx].SetDoubleRGB(atof(args[2]),atof(args[3]),atof(args[4]));
				}
			}
		}
	}
	return YSOK;
}

YSRESULT PolyCreConfig::Load(const wchar_t fn[])
{
	FILE *fp=YsFileIO::Fopen(fn,"r");
	if(NULL!=fp)
	{
		YsTextFileInputStream inStream(fp);
		Load(inStream);
		fclose(fp);
		return YSOK;
	}
	return YSERR;
}

YSRESULT PolyCreConfig::Save(YsTextOutputStream &outStream) const
{
	switch(coordSys)
	{
	case YSLEFT_ZPLUS_YPLUS:
		outStream.Printf("COORDSYS LEFT\n");
		break;
	case YSRIGHT_ZMINUS_YPLUS:
		outStream.Printf("COORDSYS RIGHT\n");
		break;
	}

	outStream.Printf("PERSPECTIVE %s\n",YsBoolToStr(perspective));

	{
		YsString utf8;
		utf8.EncodeUTF8 <wchar_t> (defDataDir);
		outStream.Printf("DATADIR \"%s\"\n",utf8.Txt());
	}

	outStream.Printf("BGCOLOR 0 %.2lf %.2lf %.2lf\n",bgColor[0].Rd(),bgColor[0].Gd(),bgColor[0].Bd());
	outStream.Printf("BGCOLOR 1 %.2lf %.2lf %.2lf\n",bgColor[1].Rd(),bgColor[1].Gd(),bgColor[1].Bd());
	outStream.Printf("BGCOLOR 2 %.2lf %.2lf %.2lf\n",bgColor[2].Rd(),bgColor[2].Gd(),bgColor[2].Bd());

	return YSOK;
}

YSRESULT PolyCreConfig::Save(const wchar_t fn[])
{
	FILE *fp=YsFileIO::Fopen(fn,"w");
	if(NULL!=fp)
	{
		YsTextFileOutputStream outStream(fp);
		Save(outStream);
		fclose(fp);
		return YSOK;
	}
	return YSERR;
}


