/* ////////////////////////////////////////////////////////////

  File Name: ysshellext.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "ysshellext.h"
#include "ysshellextmisc.h"
#include "ysshellextio.h"
#include "ysshellextsearch.h"


void YsShellExt::CommonProp::Initialize(void)
{
	ident=-1;
	label.Set("");
	refCount=0;
	refCountFromFrozenEntity=0;
}

void YsShellExt::BoundaryLayerSpec::Initialize(void)
{
	nLayer=0;
	firstLayerThickness=0.0;
	growthFuncParam=0;
	growthFuncParam=1.0;
}

void YsShellExt::ConstEdge::Initialize(void)
{
	CommonProp::Initialize();
	flags=ZEROFLAG;
	vtHdArray.Clear();
	bLayerSpec.Initialize();
	elemSize=0.0;
}

YSBOOL YsShellExt::ConstEdge::IsLoop(void) const
{
	return ((flags & LOOP) ? YSTRUE : YSFALSE);
}

void YsShellExt::ConstEdge::SetIsLoop(YSBOOL isLoop)
{
	if(YSTRUE==isLoop)
	{
		flags=(CEFLAG)(flags|LOOP);
	}
	else
	{
		flags=(CEFLAG)(flags&~LOOP);
	}
}

void YsShellExt::FaceGroup::Initialize(void)
{
	CommonProp::Initialize();
	flags=ZEROFLAG;
	plHdArray.Clear();
	bLayerSpec.Initialize();
	elemSize=0.0;
}



void YsShellExt::Volume::Initialize(void)
{
	fgHdArray.Clear();
}



void YsShellExt::VertexAttrib::Initialize(void)
{
	flags=ZEROFLAG;
	refCount=0;
	refCountFromFrozenEntity=0;
	thickness=0.0;
	oPlId=-1;
	oVtId[0]=-1;
	oVtId[1]=-1;
}

YSBOOL YsShellExt::VertexAttrib::IsRound(void) const
{
	return (0!=(flags&ROUND) ? YSTRUE : YSFALSE);
}

void YsShellExt::VertexAttrib::SetRound(YSBOOL round)
{
	if(YSTRUE==round)
	{
		flags=(VTFLAG)(flags|ROUND);
	}
	else
	{
		flags=(VTFLAG)(flags&~ROUND);
	}
}

YSBOOL YsShellExt::VertexAttrib::IsConstrained(void) const
{
	return (0!=(flags&CONSTRAINED) ? YSTRUE : YSFALSE);
}

void YsShellExt::VertexAttrib::SetConstrained(YSBOOL isConstrained)
{
	if(YSTRUE==isConstrained)
	{
		flags=(VTFLAG)(flags|CONSTRAINED);
	}
	else
	{
		flags=(VTFLAG)(flags&~CONSTRAINED);
	}
}

void YsShellExt::VertexAttrib::SetVertexTrackingToVertex(int oVtId)
{
	this->oVtId[0]=oVtId;
	this->oVtId[1]=oVtId;
}

void YsShellExt::VertexAttrib::SetVertexTrackingToEdge(int oEdVtId0,int oEdVtId1)
{
	this->oVtId[0]=oEdVtId0;
	this->oVtId[1]=oEdVtId1;
}

void YsShellExt::VertexAttrib::SetVertexTrackingToPolygon(int oPlId)
{
	this->oPlId=oPlId;
}

void YsShellExt::VertexAttrib::CopyFrom(const VertexAttrib &incoming)
{
	flags=incoming.flags;
	thickness=incoming.thickness;
}

void YsShellExt::PolygonAttrib::Initialize(void)
{
	flags=ZEROFLAG;
	refCount=0;
	refCountFromFrozenEntity=0;
	elemSize=0.0;
	thickness=0.0;
}

YSBOOL YsShellExt::PolygonAttrib::GetNoShading(void) const
{
	return (0!=(flags&NO_SHADING) ? YSTRUE : YSFALSE);
}

void YsShellExt::PolygonAttrib::SetNoShading(YSBOOL noShading)
{
	if(YSTRUE==noShading)
	{
		flags=(PLFLAG)(flags|NO_SHADING);
	}
	else
	{
		flags=(PLFLAG)(flags&~NO_SHADING);
	}
}

void YsShellExt::PolygonAttrib::SetTwoSided(YSBOOL twoSided)
{
	if(YSTRUE==twoSided)
	{
		flags=(PLFLAG)(flags|TWO_SIDED);
	}
	else
	{
		flags=(PLFLAG)(flags&~TWO_SIDED);
	}
}

void YsShellExt::PolygonAttrib::CopyFrom(const PolygonAttrib &incoming) // Copy except reference count
{
	flags=incoming.flags;
	elemSize=incoming.elemSize;
	thickness=incoming.thickness;
}



////////////////////////////////////////////////////////////



YsShellExt::YsShellExt()
{
	constEdgeArray.EnableSearch();
	faceGroupArray.EnableSearch();
	volumeArray.EnableSearch();

	shlSearch=NULL;
	extSearch=NULL;

	drawingBuffer=new YsShellExtDrawingBuffer;
}

YsShellExt::~YsShellExt()
{
	delete drawingBuffer;
	DisableSearch();
}

const YsShellExt &YsShellExt::operator=(const YsShellExt &from)
{
	CopyFrom(from);
	return *this;
}

YsShellExt::YsShellExt(const YsShellExt &from)
{
	CopyFrom(from);
}

void YsShellExt::CopyFrom(const YsShellExt &from)
{
	CleanUp();

	YsHashTable <YsShellVertexHandle> fromVtKeyToNewVtHd;
	for(auto fromVtHd : from.AllVertex())
	{
		auto newVtHd=AddVertex(from.GetVertexPosition(fromVtHd));
		auto attrib=from.GetVertexAttrib(fromVtHd);
		SetVertexAttrib(newVtHd,*attrib);

		fromVtKeyToNewVtHd.AddElement(from.GetSearchKey(fromVtHd),newVtHd);
	}

	YsHashTable <YsShellPolygonHandle> fromPlKeyToNewPlHd;
	for(auto fromPlHd : from.AllPolygon())
	{
		YsArray <YsShellVertexHandle,4> plVtHd;
		from.GetPolygon(plVtHd,fromPlHd);

		YSBOOL skip=YSFALSE;
		for(auto &vtHd : plVtHd)
		{
			if(YSOK!=fromVtKeyToNewVtHd.FindElement(vtHd,from.GetSearchKey(vtHd)))
			{
				skip=YSTRUE;
				break;
			}
		}

		if(YSTRUE!=skip)
		{
			auto newPlHd=AddPolygon(plVtHd);
			auto attrib=from.GetPolygonAttrib(fromPlHd);
			auto nom=from.GetNormal(fromPlHd);
			auto col=from.GetColor(fromPlHd);

			SetPolygonAttrib(newPlHd,*attrib);
			SetPolygonNormal(newPlHd,nom);
			SetPolygonColor(newPlHd,col);

			fromPlKeyToNewPlHd.AddElement(from.GetSearchKey(fromPlHd),newPlHd);
		}
	}

	for(auto fromCeHd : from.AllConstEdge())
	{
		YsArray <YsShellVertexHandle> ceVtHd;
		YSBOOL isLoop;
		from.GetConstEdge(ceVtHd,isLoop,fromCeHd);

		YSBOOL skip=YSFALSE;
		for(auto &vtHd : ceVtHd)
		{
			if(YSOK!=fromVtKeyToNewVtHd.FindElement(vtHd,from.GetSearchKey(vtHd)))
			{
				skip=YSTRUE;
				break;
			}
		}

		if(YSTRUE!=skip)
		{
			auto newCeHd=AddConstEdge(ceVtHd,isLoop);
			auto attrib=from.GetConstEdgeAttrib(fromCeHd);
			SetConstEdgeAttrib(newCeHd,attrib);
		}
	}

	YsHashTable <YsShellExt::FaceGroupHandle> fromFgKeyToNewFgHd;
	for(auto fromFgHd : from.AllFaceGroup())
	{
		YsArray <YsShellPolygonHandle> fgPlHd;
		from.GetFaceGroup(fgPlHd,fromFgHd);

		YSBOOL skip=YSFALSE;
		for(auto &plHd : fgPlHd)
		{
			if(YSOK!=fromPlKeyToNewPlHd.FindElement(plHd,from.GetSearchKey(plHd)))
			{
				skip=YSTRUE;
				break;
			}
		}

		if(YSTRUE!=skip)
		{
			auto newFgHd=AddFaceGroup(fgPlHd);
			auto attrib=from.GetFaceGroupAttrib(fromFgHd);
			SetFaceGroupAttrib(newFgHd,attrib);

			fromFgKeyToNewFgHd.AddElement(from.GetSearchKey(fromFgHd),newFgHd);
		}
	}

	for(auto fromVlHd : from.AllVolume())
	{
		YsArray <YsShellExt::FaceGroupHandle> vlFgHd;
		from.GetVolume(vlFgHd,fromVlHd);

		YSBOOL skip=YSFALSE;
		for(auto &fgHd : vlFgHd)
		{
			if(YSOK!=fromFgKeyToNewFgHd.FindElement(fgHd,from.GetSearchKey(fgHd)))
			{
				skip=YSTRUE;
				break;
			}
		}

		if(YSTRUE!=skip)
		{
			auto newVlHd=AddVolume(vlFgHd);
			auto attrib=from.GetVolumeAttrib(fromVlHd);
			SetVolumeAttrib(newVlHd,attrib);
		}
	}
}

YsShellExt::operator const YsShell &() const
{
	return *this;
}

void YsShellExt::CleanUp(void)
{
	YsShell::CleanUp();

	vtKeyToVtAttribArrayHandle.CleanUp();
	vtAttribArray.CleanUp();

	plKeyToPlAttribArrayHandle.CleanUp();
	plAttribArray.CleanUp();

	constEdgeArray.CleanUp();
	faceGroupArray.CleanUp();
	volumeArray.CleanUp();
}

void YsShellExt::CacheVertexNormal(void)
{
	AutoComputeVertexNormalAll(YSTRUE);
}

void YsShellExt::Encache(void) const
{
	YsShell::Encache();
	constEdgeArray.Encache();
	faceGroupArray.Encache();
	volumeArray.Encache();
}

YSRESULT YsShellExt::LoadSrf(YsTextInputStream &inStream)
{
	YsShellExtReader srfReader;
	return srfReader.MergeSrf(*this,inStream);
}

YSRESULT YsShellExt::LoadStl(const char fn[])
{
	YSRESULT res=YsShell::LoadStl(fn);
	if(YSOK==res)
	{
		for(YsShellVertexHandle vtHd=NULL; YSOK==MoveToNextVertex(vtHd); )
		{
			const YSHASHKEY vtKey=GetSearchKey(vtHd);

			YsEditArrayObjectHandle <VertexAttrib> vtAttribHd=vtAttribArray.Create();
			vtKeyToVtAttribArrayHandle.AddElement(vtKey,vtAttribHd);

			vtAttribArray[vtAttribHd]->Initialize();
		}

		for(YsShellPolygonHandle plHd=NULL; YSOK==MoveToNextPolygon(plHd); )
		{
			const YSHASHKEY plKey=GetSearchKey(plHd);

			YsEditArrayObjectHandle <PolygonAttrib> plAttribHd=plAttribArray.Create();
			plKeyToPlAttribArrayHandle.AddElement(plKey,plAttribHd);

			plAttribArray[plAttribHd]->Initialize();

			int nPlVt;
			const YsShellVertexHandle *plVtHd;
			GetVertexListOfPolygon(nPlVt,plVtHd,plHd);

			for(YSSIZE_T idx=0; idx<nPlVt; ++idx)
			{
				VertexAttrib *vtAttrib=GetVertexAttrib(plVtHd[idx]);
				++(vtAttrib->refCount);
			}
		}
	}
	return res;
}

YSRESULT YsShellExt::LoadObj(YsTextInputStream &inStream)
{
	YsShellExtObjReader::ReadOption opt;
	YsShellExtObjReader objReader;
	return objReader.ReadObj(*this,inStream,opt);
}

YsShellExtDrawingBuffer &YsShellExt::GetDrawingBuffer(void) const
{
	return *drawingBuffer;
}

YsShellVertexHandle YsShellExt::AddVertex(const YsVec3 &pos)
{
	const YsShellVertexHandle newVtHd=YsShell::AddVertexH(pos);
	const YSHASHKEY vtKey=GetSearchKey(newVtHd);

	YsEditArrayObjectHandle <VertexAttrib> vtAttribHd=vtAttribArray.Create();
	vtKeyToVtAttribArrayHandle.AddElement(vtKey,vtAttribHd);

	vtAttribArray[vtAttribHd]->Initialize();

	return newVtHd;
}

YSRESULT YsShellExt::SetVertexRoundFlag(YsShellVertexHandle vtHd,YSBOOL round)
{
	VertexAttrib *vtAttrib=GetVertexAttrib(vtHd);
	if(NULL!=vtAttrib)
	{
		vtAttrib->SetRound(round);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::MoveToNextVertex(YsShellVertexHandle &vtHd) const
{
	vtHd=FindNextVertex(vtHd);
	if(NULL!=vtHd)
	{
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsShellExt::MoveToNextPolygon(YsShellPolygonHandle &plHd) const
{
	plHd=FindNextPolygon(plHd);
	if(NULL!=plHd)
	{
		return YSOK;
	}
	else
	{
		return YSERR;
	}
}

YSRESULT YsShellExt::SetVertexConstrainedFlag(YsShellVertexHandle vtHd,YSBOOL round)
{
	VertexAttrib *vtAttrib=GetVertexAttrib(vtHd);
	if(NULL!=vtAttrib)
	{
		vtAttrib->SetConstrained(round);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::FreezeVertex(YsShellVertexHandle vtHd)
{
	VertexAttrib *vtAttrib=GetVertexAttrib(vtHd);
	if(NULL!=vtAttrib && 0==vtAttrib->refCount)
	{
		return YsShell::FreezeVertex(vtHd);
	}
	return YSERR;
}

YSRESULT YsShellExt::MeltVertex(YsShellVertexHandle vtHd)
{
	return YsShell::MeltVertex(vtHd);
}

YSRESULT YsShellExt::DeleteVertex(YsShellVertexHandle vtHd)
{
	VertexAttrib *vtAttrib=GetVertexAttrib(vtHd);
	if(NULL!=vtAttrib && YSTRUE!=IsFrozen(vtHd) && 0==vtAttrib->refCount && 0==vtAttrib->refCountFromFrozenEntity)
	{
		return YsShell::DeleteVertex(vtHd);
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteFrozenVertex(YsShellVertexHandle vtHd)
{
	YsShellExt::VertexAttrib *attrib=GetVertexAttribMaybeFrozen(vtHd);
	if(NULL!=attrib && YSTRUE==IsFrozen(vtHd) && 0==attrib->refCount && 0==attrib->refCountFromFrozenEntity)
	{
		YsShell::DeleteFrozenVertex(vtHd);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetVertexTrackingToVertex(YsShellVertexHandle vtHd,int oVtId)
{
	YsShellExt::VertexAttrib *vtAttrib=GetVertexAttrib(vtHd);
	if(NULL!=vtAttrib)
	{
		vtAttrib->SetVertexTrackingToVertex(oVtId);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetVertexTrackingToEdge(YsShellVertexHandle vtHd,int oEdVtId0,int oEdVtId1)
{
	YsShellExt::VertexAttrib *vtAttrib=GetVertexAttrib(vtHd);
	if(NULL!=vtAttrib)
	{
		vtAttrib->SetVertexTrackingToEdge(oEdVtId0,oEdVtId1);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetVertexTrackingToPolygon(YsShellVertexHandle vtHd,int oPlId)
{
	YsShellExt::VertexAttrib *vtAttrib=GetVertexAttrib(vtHd);
	if(NULL!=vtAttrib)
	{
		vtAttrib->SetVertexTrackingToPolygon(oPlId);
		return YSOK;
	}
	return YSERR;
}

YsShellPolygonHandle YsShellExt::AddPolygon(YSSIZE_T nVt,const YsShellVertexHandle vtHdArray[])
{
	const YsShellPolygonHandle newPlHd=YsShell::AddPolygonH(nVt,vtHdArray);
	const YSHASHKEY plKey=GetSearchKey(newPlHd);

	YsEditArrayObjectHandle <PolygonAttrib> plAttribHd=plAttribArray.Create();
	plKeyToPlAttribArrayHandle.AddElement(plKey,plAttribHd);

	plAttribArray[plAttribHd]->Initialize();

	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		VertexAttrib *vtAttrib=GetVertexAttrib(vtHdArray[idx]);
		++(vtAttrib->refCount);
	}

	return newPlHd;
}

YSRESULT YsShellExt::SetPolygonVertex(YsShellPolygonHandle plHd,YSSIZE_T nVt,const YsShellVertexHandle vtHdArray[])
{
	YsArray <YsShellVertexHandle,16> oldPlVtHd;
	GetVertexListOfPolygon(oldPlVtHd,plHd);
	for(YSSIZE_T idx=0; idx<oldPlVtHd.GetN(); ++idx)
	{
		VertexAttrib *vtAttrib=GetVertexAttrib(oldPlVtHd[idx]);
		--(vtAttrib->refCount);
	}

	for(YSSIZE_T idx=0; idx<nVt; ++idx)
	{
		VertexAttrib *vtAttrib=GetVertexAttrib(vtHdArray[idx]);
		++(vtAttrib->refCount);
	}

	return YsShell::ModifyPolygon(plHd,nVt,vtHdArray);
}

YSRESULT YsShellExt::SetPolygonNoShadingFlag(YsShellPolygonHandle plHd,YSBOOL noShading)
{
	PolygonAttrib *plAttrib=GetPolygonAttrib(plHd);
	if(NULL!=plAttrib)
	{
		plAttrib->SetNoShading(noShading);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetPolygonTwoSidedFlag(YsShellPolygonHandle plHd,YSBOOL twoSided)
{
	PolygonAttrib *plAttrib=GetPolygonAttrib(plHd);
	if(NULL!=plAttrib)
	{
		plAttrib->SetTwoSided(twoSided);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetPolygonDesiredElementSize(YsShellPolygonHandle plHd,const double elemSize)
{
	PolygonAttrib *plAttrib=GetPolygonAttrib(plHd);
	if(NULL!=plAttrib)
	{
		plAttrib->elemSize=elemSize;
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetPolygonColor(YsShellPolygonHandle plHd,const YsColor &col)
{
	return YsShell::SetColorOfPolygon(plHd,col);
}

YSRESULT YsShellExt::SetPolygonNormal(YsShellPolygonHandle plHd,const YsVec3 &nom)
{
	return YsShell::SetNormalOfPolygon(plHd,nom);
}

YSRESULT YsShellExt::SetPolygonAlpha(YsShellPolygonHandle plHd,const double alpha)
{
	if(NULL!=plHd)
	{
		YsColor col;
		GetColorOfPolygon(col,plHd);
		col.SetAd(alpha);
		SetColorOfPolygon(plHd,col);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::FreezePolygon(YsShellPolygonHandle plHd)
{
	PolygonAttrib *plAttrib=GetPolygonAttrib(plHd);
	if(NULL!=plAttrib && YSTRUE!=YsShell::IsFrozen(plHd) && 0==plAttrib->refCount)
	{
		int nVt;
		const YsShellVertexHandle *plVtHd;
		GetVertexListOfPolygon(nVt,plVtHd,plHd);

		for(YSSIZE_T idx=0; idx<nVt; ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(plVtHd[idx]);
			--vtAttrib->refCount;
			++vtAttrib->refCountFromFrozenEntity;
		}

		return YsShell::FreezePolygon(plHd);
	}
	return YSERR;
}

YSRESULT YsShellExt::MeltPolygon(YsShellPolygonHandle plHd)
{
	if(YSTRUE==YsShell::IsFrozen(plHd))
	{
		YsShell::MeltPolygon(plHd);

		int nVt;
		const YsShellVertexHandle *plVtHd;
		GetVertexListOfPolygon(nVt,plVtHd,plHd);

		for(YSSIZE_T idx=0; idx<nVt; ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(plVtHd[idx]);
			++vtAttrib->refCount;
			--vtAttrib->refCountFromFrozenEntity;
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::DeletePolygon(YsShellPolygonHandle plHd)
{
	PolygonAttrib *plAttrib=GetPolygonAttrib(plHd);
	if(NULL!=plAttrib && YSTRUE!=IsFrozen(plHd) && 0==plAttrib->refCount && 0==plAttrib->refCountFromFrozenEntity)
	{
		int nVt;
		const YsShellVertexHandle *plVtHd;
		GetVertexListOfPolygon(nVt,plVtHd,plHd);

		for(YSSIZE_T idx=0; idx<nVt; ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(plVtHd[idx]);
			--vtAttrib->refCount;
		}

		return YsShell::DeletePolygon(plHd);
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteFrozenPolygon(YsShellPolygonHandle plHd)
{
	PolygonAttrib *plAttrib=GetPolygonAttribMaybeFrozen(plHd);
	if(NULL!=plAttrib && YSTRUE==IsFrozen(plHd) && 0==plAttrib->refCount && 0==plAttrib->refCountFromFrozenEntity)
	{
		int nVt;
		const YsShellVertexHandle *plVtHd;
		GetPolygonMaybeFrozen(nVt,plVtHd,plHd);

		for(YSSIZE_T idx=0; idx<nVt; ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttribMaybeFrozen(plVtHd[idx]);
			--vtAttrib->refCountFromFrozenEntity;
		}

		return YsShell::DeleteFrozenPolygon(plHd);
	}
	return YSERR;
}

const YsShellExt::VertexAttrib *YsShellExt::GetVertexAttrib(YsShellVertexHandle vtHd) const
{
	YsEditArrayObjectHandle <VertexAttrib> vtAttribHd;
	if(YSOK==vtKeyToVtAttribArrayHandle.FindElement(vtAttribHd,GetSearchKey(vtHd)))
	{
		return vtAttribArray[vtAttribHd];
	}
	return NULL;
}

YsShellExt::VertexAttrib *YsShellExt::GetVertexAttrib(YsShellVertexHandle vtHd)
{
	YsEditArrayObjectHandle <VertexAttrib> vtAttribHd;
	if(YSOK==vtKeyToVtAttribArrayHandle.FindElement(vtAttribHd,GetSearchKey(vtHd)))
	{
		return vtAttribArray[vtAttribHd];
	}
	return NULL;
}

YsShellExt::VertexAttrib *YsShellExt::GetVertexAttribMaybeFrozen(YsShellVertexHandle vtHd)
{
	YsEditArrayObjectHandle <VertexAttrib> vtAttribHd;
	if(YSOK==vtKeyToVtAttribArrayHandle.FindElement(vtAttribHd,GetSearchKeyMaybeFrozen(vtHd)))
	{
		return vtAttribArray.GetObjectMaybeFrozen(vtAttribHd);
	}
	return NULL;
}

const YsShellExt::VertexAttrib *YsShellExt::GetVertexAttribMaybeFrozen(YsShellVertexHandle vtHd) const
{
	YsEditArrayObjectHandle <VertexAttrib> vtAttribHd;
	if(YSOK==vtKeyToVtAttribArrayHandle.FindElement(vtAttribHd,GetSearchKeyMaybeFrozen(vtHd)))
	{
		return vtAttribArray.GetObjectMaybeFrozen(vtAttribHd);
	}
	return NULL;
}

YSRESULT YsShellExt::SetVertexAttrib(YsShellVertexHandle vtHd,const VertexAttrib &incoming)
{
	YsEditArrayObjectHandle <VertexAttrib> vtAttribHd;
	if(YSOK==vtKeyToVtAttribArrayHandle.FindElement(vtAttribHd,GetSearchKey(vtHd)))
	{
		vtAttribArray[vtAttribHd]->CopyFrom(incoming);
		return YSOK;
	}
	return YSERR;
}

YSBOOL YsShellExt::IsVertexUsed(YsShellVertexHandle vtHd) const
{
	YsEditArrayObjectHandle <VertexAttrib> vtAttribHd;
	if(YSOK==vtKeyToVtAttribArrayHandle.FindElement(vtAttribHd,GetSearchKey(vtHd)) &&
	   0<vtAttribArray[vtAttribHd]->refCount)
	{
		return YSTRUE;
	}
	return YSFALSE;
}

const YsShellExt::PolygonAttrib *YsShellExt::GetPolygonAttrib(YsShellPolygonHandle plHd) const
{
	YsEditArrayObjectHandle <PolygonAttrib> plAttribHd;
	if(YSOK==plKeyToPlAttribArrayHandle.FindElement(plAttribHd,GetSearchKey(plHd)))
	{
		return plAttribArray[plAttribHd];
	}
	return NULL;
}

YsShellExt::PolygonAttrib *YsShellExt::GetPolygonAttrib(YsShellPolygonHandle plHd)
{
	YsEditArrayObjectHandle <PolygonAttrib> plAttribHd;
	if(YSOK==plKeyToPlAttribArrayHandle.FindElement(plAttribHd,GetSearchKey(plHd)))
	{
		return plAttribArray[plAttribHd];
	}
	return NULL;
}

YsShellExt::PolygonAttrib *YsShellExt::GetPolygonAttribMaybeFrozen(YsShellPolygonHandle plHd)
{
	YsEditArrayObjectHandle <PolygonAttrib> plAttribHd;
	if(YSOK==plKeyToPlAttribArrayHandle.FindElement(plAttribHd,GetSearchKeyMaybeFrozen(plHd)))
	{
		return plAttribArray.GetObjectMaybeFrozen(plAttribHd);
	}
	return NULL;
}

const YsShellExt::PolygonAttrib *YsShellExt::GetPolygonAttribMaybeFrozen(YsShellPolygonHandle plHd) const
{
	YsEditArrayObjectHandle <PolygonAttrib> plAttribHd;
	if(YSOK==plKeyToPlAttribArrayHandle.FindElement(plAttribHd,GetSearchKeyMaybeFrozen(plHd)))
	{
		return plAttribArray.GetObjectMaybeFrozen(plAttribHd);
	}
	return NULL;
}

YSRESULT YsShellExt::SetPolygonAttrib(YsShellPolygonHandle plHd,const PolygonAttrib &inAttrib)
{
	PolygonAttrib *outAttrib=GetPolygonAttrib(plHd);
	outAttrib->CopyFrom(inAttrib);

	return YSOK;
}

YSSIZE_T YsShellExt::GetNumConstEdge(void) const
{
	return constEdgeArray.GetN();
}

YSHASHKEY YsShellExt::GetSearchKey(ConstEdgeHandle ceHd) const
{
	return constEdgeArray.GetSearchKey(ceHd);
}

YSRESULT YsShellExt::MoveToNextConstEdge(ConstEdgeHandle &ceHd) const
{
	return constEdgeArray.MoveToNext(ceHd);
}

YSRESULT YsShellExt::MoveToPrevConstEdge(ConstEdgeHandle &ceHd) const
{
	return constEdgeArray.MoveToPrev(ceHd);
}

YSRESULT YsShellExt::GetConstEdge(YSSIZE_T &nCeVt,const YsShellVertexHandle *&ceVtHd,YSBOOL &isLoop,ConstEdgeHandle ceHd) const
{
	const ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp)
	{
		isLoop=ceProp->IsLoop();
		nCeVt=ceProp->vtHdArray.GetN();
		ceVtHd=ceProp->vtHdArray;
		return YSOK;
	}
	nCeVt=0;
	ceVtHd=NULL;
	isLoop=YSFALSE;
	return YSERR;
}

YSBOOL YsShellExt::IsConstEdgeLoop(ConstEdgeHandle ceHd) const
{
	const ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp)
	{
		return ceProp->IsLoop();
	}
	return YSTFUNKNOWN;
}

YsShellExt::ConstEdgeHandle YsShellExt::AddConstEdge(YSSIZE_T nVt,const YsShellVertexHandle ceVtHd[],YSBOOL isLoop)
{
	ConstEdgeHandle newCeHd=constEdgeArray.Create();
	if(NULL!=newCeHd)
	{
		ConstEdge *ceProp=constEdgeArray[newCeHd];
		ceProp->Initialize();
		ceProp->SetIsLoop(isLoop);
		ceProp->vtHdArray.Set(nVt,ceVtHd);

		for(YSSIZE_T idx=0; idx<nVt; ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(ceVtHd[idx]);
			++vtAttrib->refCount;
		}

		extSearch->ConstEdgeAdded(newCeHd);
	}
	return newCeHd;
}

YsShellExt::ConstEdge YsShellExt::GetConstEdgeAttrib(YsShellExt::ConstEdgeHandle ceHd) const
{
	ConstEdge attrib;
	const ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp)
	{
		attrib=(*ceProp);
		attrib.vtHdArray.CleanUp();
	}
	return attrib;
}

YSRESULT YsShellExt::SetConstEdgeAttrib(YsShellExt::ConstEdgeHandle ceHd,const YsShellExt::ConstEdge &attrib)
{
	ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp)
	{
		const int ident=ceProp->ident;
		const int refCount=ceProp->refCount;
		const int refCountFromFrozenEntity=ceProp->refCountFromFrozenEntity;
		YsArray <YsShellVertexHandle> vtHdArray;
		vtHdArray.MoveFrom(ceProp->vtHdArray);

		const YSBOOL isLoop=ceProp->IsLoop();
		*ceProp=attrib;
		ceProp->SetIsLoop(isLoop);

		ceProp->ident=ident;
		ceProp->refCount=refCount;
		ceProp->refCountFromFrozenEntity=refCountFromFrozenEntity;
		ceProp->vtHdArray.MoveFrom(vtHdArray);

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetConstEdgeIsLoop(ConstEdgeHandle ceHd,YSBOOL isLoop)
{
	ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd))
	{
		ceProp->SetIsLoop(isLoop);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::ModifyConstEdge(ConstEdgeHandle ceHd,YSSIZE_T nVt,const YsShellVertexHandle ceVtHd[],YSBOOL isLoop)
{
	ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd))
	{
		extSearch->ConstEdgeAboutToBeDeleted(ceHd);

		for(YSSIZE_T idx=0; idx<ceProp->vtHdArray.GetN(); ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(ceProp->vtHdArray[idx]);
			--vtAttrib->refCount;
		}

		ceProp->vtHdArray.Set(nVt,ceVtHd);
		if(YSTFUNKNOWN!=isLoop)
		{
			ceProp->SetIsLoop(isLoop);
		}

		for(YSSIZE_T idx=0; idx<ceProp->vtHdArray.GetN(); ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(ceProp->vtHdArray[idx]);
			++vtAttrib->refCount;
		}

		extSearch->ConstEdgeAdded(ceHd);

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::AddConstEdgeVertex(ConstEdgeHandle ceHd,YSSIZE_T nVt,const YsShellVertexHandle ceVtHd[])
{
	ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd))
	{
		YSSIZE_T idx0=ceProp->vtHdArray.GetN();
		ceProp->vtHdArray.Append(nVt,ceVtHd);
		for(YSSIZE_T idx=idx0; idx<ceProp->vtHdArray.GetN(); ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(ceProp->vtHdArray[idx]);
			++vtAttrib->refCount;
		}

		extSearch->ConstEdgeVertexAdded(ceHd,nVt,ceVtHd);

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetConstEdgeLabel(ConstEdgeHandle ceHd,const char label[])
{
	ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd))
	{
		ceProp->label=label;
		return YSOK;
	}
	return YSERR;
}

const char *YsShellExt::GetConstEdgeLabel(ConstEdgeHandle ceHd) const
{
	const ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd))
	{
		return ceProp->label;
	}
	return "";
}

YSRESULT YsShellExt::SetConstEdgeIdent(ConstEdgeHandle ceHd,YSSIZE_T ident)
{
	ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd))
	{
		ceProp->ident=ident;
		return YSOK;
	}
	return YSERR;
}

YSSIZE_T YsShellExt::GetConstEdgeIdent(ConstEdgeHandle ceHd) const
{
	const ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd))
	{
		return ceProp->ident;
	}
	return -1;
}

YSRESULT YsShellExt::FreezeConstEdge(YsShellExt::ConstEdgeHandle ceHd)
{
	const ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && 0==ceProp->refCount)
	{
		extSearch->ConstEdgeAboutToBeDeleted(ceHd);

		for(YSSIZE_T idx=0; idx<ceProp->vtHdArray.GetN(); ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(ceProp->vtHdArray[idx]);
			--vtAttrib->refCount;
			++vtAttrib->refCountFromFrozenEntity;
		}

		constEdgeArray.Freeze(ceHd);

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::MeltConstEdge(ConstEdgeHandle ceHd)
{
	if(YSTRUE==constEdgeArray.IsFrozen(ceHd))
	{
		constEdgeArray.Melt(ceHd);

		const ConstEdge *ceProp=constEdgeArray[ceHd];
		for(YSSIZE_T idx=0; idx<ceProp->vtHdArray.GetN(); ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(ceProp->vtHdArray[idx]);
			++vtAttrib->refCount;
			--vtAttrib->refCountFromFrozenEntity;
		}

		extSearch->ConstEdgeAdded(ceHd);
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteConstEdge(YsShellExt::ConstEdgeHandle ceHd)
{
	ConstEdge *ceProp=constEdgeArray[ceHd];
	if(NULL!=ceProp && YSTRUE!=constEdgeArray.IsFrozen(ceHd) && 0==ceProp->refCount && 0==ceProp->refCountFromFrozenEntity)
	{
		extSearch->ConstEdgeAboutToBeDeleted(ceHd);

		for(YSSIZE_T idx=0; idx<ceProp->vtHdArray.GetN(); ++idx)
		{
			VertexAttrib *vtAttrib=GetVertexAttrib(ceProp->vtHdArray[idx]);
			--vtAttrib->refCount;
		}

		constEdgeArray.Delete(ceHd);

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteFrozenConstEdge(YsShellExt::ConstEdgeHandle ceHd)
{
	if(YSTRUE==constEdgeArray.IsFrozen(ceHd))
	{
		auto ceProp=constEdgeArray.GetFrozenObject(ceHd);

		if(NULL!=ceProp && 0==ceProp->refCount && 0==ceProp->refCountFromFrozenEntity)
		{
			for(YSSIZE_T idx=0; idx<ceProp->vtHdArray.GetN(); ++idx)
			{
				VertexAttrib *vtAttrib=GetVertexAttribMaybeFrozen(ceProp->vtHdArray[idx]);
				--vtAttrib->refCountFromFrozenEntity;
			}

			constEdgeArray.DeleteFrozen(ceHd);
			return YSOK;
		}
	}
	return YSERR;
}

YSSIZE_T YsShellExt::GetNumFaceGroup(void) const
{
	return faceGroupArray.GetN();
}

YSHASHKEY YsShellExt::GetSearchKey(FaceGroupHandle fgHd) const
{
	return faceGroupArray.GetSearchKey(fgHd);
}

YSRESULT YsShellExt::MoveToNextFaceGroup(FaceGroupHandle &fgHd) const
{
	return faceGroupArray.MoveToNext(fgHd);
}

YSRESULT YsShellExt::MoveToPrevFaceGroup(FaceGroupHandle &fgHd) const
{
	return faceGroupArray.MoveToPrev(fgHd);
}

YSRESULT YsShellExt::GetFaceGroup(YSSIZE_T &nFgPl,const YsShellPolygonHandle *&fgPlHd,FaceGroupHandle fgHd) const
{
	const FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp)
	{
		nFgPl=fgProp->plHdArray.GetN();
		fgPlHd=fgProp->plHdArray;
		return YSOK;
	}
	nFgPl=0;
	fgPlHd=NULL;
	return YSERR;
}

YsArray <YsShellPolygonHandle> YsShellExt::GetFaceGroup(FaceGroupHandle fgHd) const
{
	YsArray <YsShellPolygonHandle> fgPlHd;
	GetFaceGroup(fgPlHd,fgHd);
	return fgPlHd;
}

YsShellExt::FaceGroup YsShellExt::GetFaceGroupAttrib(FaceGroupHandle fgHd) const
{
	FaceGroup attrib;
	auto fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp)
	{
		attrib=(*fgProp);
		attrib.plHdArray.CleanUp();
	}
	return attrib;
}

YSRESULT YsShellExt::SetFaceGroupAttrib(YsShellExt::FaceGroupHandle fgHd,const FaceGroup &attrib)
{
	auto fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp)
	{
		const int ident=fgProp->ident;
		const int refCount=fgProp->refCount;
		const int refCountFromFrozenEntity=fgProp->refCountFromFrozenEntity;
		YsArray <YsShellPolygonHandle> plHdArray;
		plHdArray.MoveFrom(fgProp->plHdArray);

		*fgProp=attrib;

		fgProp->ident=ident;
		fgProp->refCount=refCount;
		fgProp->refCountFromFrozenEntity=refCountFromFrozenEntity;
		fgProp->plHdArray.MoveFrom(plHdArray);

		return YSOK;
	}
	return YSERR;
}

YsShellExt::FaceGroupHandle YsShellExt::AddFaceGroup(YSSIZE_T nPl,const YsShellPolygonHandle fgPlHd[])
{
	FaceGroupHandle newFgHd=faceGroupArray.Create();
	if(NULL!=newFgHd)
	{
		// Should check if fgPlHd[idx] is already used by a face group.
		if(NULL!=extSearch)
		{
			for(YSSIZE_T idx=0; idx<nPl; ++idx)
			{
				if(NULL!=extSearch->FindFaceGroupFromPolygon(fgPlHd[idx]))
				{
					return NULL;
				}
			}
		}

		FaceGroup *fgProp=faceGroupArray[newFgHd];
		fgProp->Initialize();
		fgProp->plHdArray.Set(nPl,fgPlHd);

		for(YSSIZE_T idx=0; idx<nPl; ++idx)
		{
			PolygonAttrib *plAttrib=GetPolygonAttrib(fgPlHd[idx]);
			++plAttrib->refCount;
		}

		if(NULL!=extSearch)
		{
			extSearch->FaceGroupAdded(newFgHd);
		}

		return newFgHd;
	}
	return NULL;
}

YSRESULT YsShellExt::ModifyFaceGroup(FaceGroupHandle fgHd,const YSSIZE_T nPl,const YsShellPolygonHandle fgPlHd[])
{
	FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		if(NULL!=extSearch)
		{
			for(YSSIZE_T idx=0; idx<nPl; ++idx)
			{
				YsShellExt::FaceGroupHandle usingFgHd=extSearch->FindFaceGroupFromPolygon(fgPlHd[idx]);
				if(NULL!=usingFgHd && usingFgHd!=fgHd)
				{
					return YSERR;
				}
			}

			extSearch->FaceGroupAboutToBeDeleted(fgHd);
		}

		for(YSSIZE_T idx=0; idx<fgProp->plHdArray.GetN(); ++idx)
		{
			PolygonAttrib *plAttrib=GetPolygonAttrib(fgProp->plHdArray[idx]);
			--plAttrib->refCount;
		}

		fgProp->plHdArray.Set(nPl,fgPlHd);

		for(YSSIZE_T idx=0; idx<fgProp->plHdArray.GetN(); ++idx)
		{
			PolygonAttrib *plAttrib=GetPolygonAttrib(fgProp->plHdArray[idx]);
			++plAttrib->refCount;
		}

		if(NULL!=extSearch)
		{
			extSearch->FaceGroupAdded(fgHd);
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::AddFaceGroupPolygon(FaceGroupHandle fgHd,const YSSIZE_T nPl,const YsShellPolygonHandle fgPlHd[])
{
	FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		if(NULL!=extSearch)
		{
			for(YSSIZE_T idx=0; idx<nPl; ++idx)
			{
				YsShellExt::FaceGroupHandle usingFgHd=extSearch->FindFaceGroupFromPolygon(fgPlHd[idx]);
				if(NULL!=usingFgHd)
				{
					return YSERR;
				}
			}
		}

		YSSIZE_T idx0=fgProp->plHdArray.GetN();
		fgProp->plHdArray.Append(nPl,fgPlHd);
		for(YSSIZE_T idx=idx0; idx<fgProp->plHdArray.GetN(); ++idx)
		{
			PolygonAttrib *plAttrib=GetPolygonAttrib(fgProp->plHdArray[idx]);
			++plAttrib->refCount;
		}

		if(NULL!=extSearch)
		{
			extSearch->MultiplePolygonAddedToFaceGroup(fgHd,nPl,fgPlHd);
		}

		return YSOK;
	}
	return YSERR;
}

YsShellExt::FaceGroupHandle YsShellExt::DeleteFaceGroupPolygon(YsShellPolygonHandle plHd)
{
	PolygonAttrib *plAttrib=GetPolygonAttrib(plHd);
	if(NULL!=plAttrib && YSTRUE!=IsFrozen(plHd) && 0<plAttrib->refCount)
	{
		YsShellExt::FaceGroupHandle fgHd=FindFaceGroupFromPolygon(plHd);
		if(NULL!=fgHd)
		{
			YSSIZE_T pos=-1;

			for(YSSIZE_T idx=0; idx<faceGroupArray[fgHd]->plHdArray.GetN(); ++idx)
			{
				if(plHd==faceGroupArray[fgHd]->plHdArray[idx])
				{
					pos=idx;
					break;
				}
			}

			if(0<=pos)
			{
				faceGroupArray[fgHd]->plHdArray.DeleteBySwapping(pos);
				--(plAttrib->refCount);
				if(NULL!=extSearch)
				{
					extSearch->PolygonRemovedFromFaceGroup(fgHd,plHd);
				}
				return fgHd;
			}
		}
	}
	return NULL;
}

YSRESULT YsShellExt::SetFaceGroupInteriorConstSurfaceFlag(FaceGroupHandle fgHd,YSBOOL isInteriorConstSurface)
{
	FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		unsigned int flags=(unsigned int)fgProp->flags;
		if(YSTRUE==isInteriorConstSurface)
		{
			flags|=FaceGroup::INTERIOR_CONST_SURFACE;
		}
		else
		{
			flags&=~FaceGroup::INTERIOR_CONST_SURFACE;
		}
		fgProp->flags=(FaceGroup::FGFLAG)flags;
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::SetFaceGroupLabel(FaceGroupHandle fgHd,const char label[])
{
	FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		fgProp->label=label;
		return YSOK;
	}
	return YSERR;
}

const char *YsShellExt::GetFaceGroupLabel(FaceGroupHandle fgHd) const
{
	const FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		return fgProp->label;
	}
	return "";
}

YSRESULT YsShellExt::SetFaceGroupIdent(FaceGroupHandle fgHd,YSSIZE_T ident)
{
	FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		fgProp->ident=ident;
		return YSOK;
	}
	return YSERR;
}

YSSIZE_T YsShellExt::GetFaceGroupIdent(FaceGroupHandle fgHd) const
{
	const FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		return fgProp->ident;
	}
	return -1;
}

YSBOOL YsShellExt::IsInteriorConstSurface(FaceGroupHandle fgHd) const
{
	const FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp &&
	   YSTRUE!=faceGroupArray.IsFrozen(fgHd) &&
	   0!=(((unsigned int)fgProp->flags)&FaceGroup::INTERIOR_CONST_SURFACE))
	{
		return YSFALSE;
	}
	return YSFALSE;
}

YSRESULT YsShellExt::FreezeFaceGroup(YsShellExt::FaceGroupHandle fgHd)
{
	FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		if(NULL!=extSearch)
		{
			extSearch->FaceGroupAboutToBeDeleted(fgHd);
		}

		for(YSSIZE_T idx=0; idx<fgProp->plHdArray.GetN(); ++idx)
		{
			PolygonAttrib *plAttrib=GetPolygonAttrib(fgProp->plHdArray[idx]);
			--plAttrib->refCount;
			++plAttrib->refCountFromFrozenEntity;
		}

		faceGroupArray.Freeze(fgHd);

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::MeltFaceGroup(YsShellExt::FaceGroupHandle fgHd)
{
	if(YSTRUE==faceGroupArray.IsFrozen(fgHd))
	{
		faceGroupArray.Melt(fgHd);

		FaceGroup *fgProp=faceGroupArray[fgHd];
		for(YSSIZE_T idx=0; idx<fgProp->plHdArray.GetN(); ++idx)
		{
			PolygonAttrib *plAttrib=GetPolygonAttrib(fgProp->plHdArray[idx]);
			++plAttrib->refCount;
			--plAttrib->refCountFromFrozenEntity;
		}

		if(NULL!=extSearch)
		{
			extSearch->FaceGroupAdded(fgHd);
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteFaceGroup(YsShellExt::FaceGroupHandle fgHd)
{
	FaceGroup *fgProp=faceGroupArray[fgHd];
	if(NULL!=fgProp && YSTRUE!=faceGroupArray.IsFrozen(fgHd))
	{
		if(NULL!=extSearch)
		{
			extSearch->FaceGroupAboutToBeDeleted(fgHd);
		}

		for(YSSIZE_T idx=0; idx<fgProp->plHdArray.GetN(); ++idx)
		{
			PolygonAttrib *plAttrib=GetPolygonAttrib(fgProp->plHdArray[idx]);
			--plAttrib->refCount;
		}

		faceGroupArray.Delete(fgHd);

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteFrozenFaceGroup(FaceGroupHandle fgHd)
{
	if(YSTRUE==faceGroupArray.IsFrozen(fgHd))
	{
		const FaceGroup *fgProp=faceGroupArray.GetFrozenObject(fgHd);
		if(NULL!=fgProp)
		{
			for(YSSIZE_T idx=0; idx<fgProp->plHdArray.GetN(); ++idx)
			{
				PolygonAttrib *plAttrib=GetPolygonAttribMaybeFrozen(fgProp->plHdArray[idx]);
				--plAttrib->refCountFromFrozenEntity;
			}
			faceGroupArray.DeleteFrozen(fgHd);
			return YSOK;
		}
	}
	return YSERR;

}

YSBOOL YsShellExt::IsEdgeOnFaceGroupBoundary(YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1) const
{
	auto edPlHd=FindPolygonFromEdgePiece(edVtHd0,edVtHd1);
	if(0<edPlHd.GetN())
	{
		FaceGroupHandle fgHd0=NULL;
		for(auto plHd : edPlHd)
		{
			FaceGroupHandle fgHd1=FindFaceGroupFromPolygon(plHd);
			if(NULL==fgHd0)
			{
				fgHd0=fgHd1;
			}
			else
			{
				
				if(NULL!=fgHd1 && fgHd0!=fgHd1)
				{
					return YSTRUE;
				}
			}
		}

		if(1==edPlHd.GetN() && NULL!=fgHd0)
		{
			return YSTRUE;
		}
	}
	return YSFALSE;
}

YSBOOL YsShellExt::IsEdgeOnFaceGroupBoundary(YsShellVertexHandle edVtHd[2]) const
{
	return IsEdgeOnFaceGroupBoundary(edVtHd[0],edVtHd[1]);
}

YSRESULT YsShellExt::MoveToNextVolume(VolumeHandle &vlHd) const
{
	return volumeArray.MoveToNext(vlHd);
}

YSRESULT YsShellExt::MoveToPrevVolume(VolumeHandle &vlHd) const
{
	return volumeArray.MoveToPrev(vlHd);
}

YSSIZE_T YsShellExt::GetNumVolume(void) const
{
	return volumeArray.GetN();
}

YSHASHKEY YsShellExt::GetSearchKey(VolumeHandle vlHd) const
{
	return volumeArray.GetSearchKey(vlHd);
}

YSRESULT YsShellExt::GetVolume(YSSIZE_T &nVlFg,const YsShellExt::FaceGroupHandle *&vlFgHd,YsShellExt::VolumeHandle vlHd) const
{
	const Volume *vlProp=volumeArray[vlHd];
	if(NULL!=vlProp)
	{
		nVlFg=vlProp->fgHdArray.GetN();
		vlFgHd=vlProp->fgHdArray;
		return YSOK;
	}
	nVlFg=0;
	vlFgHd=NULL;
	return YSERR;
}

YsShellExt::Volume YsShellExt::GetVolumeAttrib(YsShellExt::VolumeHandle vlHd) const
{
	Volume attrib;
	const Volume *vlProp=volumeArray[vlHd];
	if(NULL!=vlProp)
	{
		attrib=(*vlProp);
		attrib.fgHdArray.CleanUp();
	}
	return attrib;
}

YSRESULT YsShellExt::SetVolumeAttrib(YsShellExt::VolumeHandle vlHd,const YsShellExt::Volume &attrib)
{
	auto vlProp=volumeArray[vlHd];
	if(NULL!=vlProp)
	{
		const int ident=vlProp->ident;
		const int refCount=vlProp->refCount;
		const int refCountFromFrozenEntity=vlProp->refCountFromFrozenEntity;
		YsArray <FaceGroupHandle> fgHdArray;
		fgHdArray.MoveFrom(vlProp->fgHdArray);

		*vlProp=attrib;

		vlProp->ident=ident;
		vlProp->refCount=refCount;
		vlProp->refCountFromFrozenEntity=refCountFromFrozenEntity;
		vlProp->fgHdArray.MoveFrom(fgHdArray);

		return YSOK;
	}
	return YSERR;
}

YsShellExt::VolumeHandle YsShellExt::AddVolume(YSSIZE_T nVlFg,const YsShellExt::FaceGroupHandle vlFgHd[])
{
	VolumeHandle newVlHd=volumeArray.Create();
	if(NULL!=newVlHd)
	{
		Volume *vlProp=volumeArray[newVlHd];
		vlProp->Initialize();
		vlProp->fgHdArray.Set(nVlFg,vlFgHd);

		for(YSSIZE_T idx=0; idx<nVlFg; ++idx)
		{
			FaceGroup *fgProp=faceGroupArray[vlFgHd[idx]];
			++fgProp->refCount;
		}

		if(NULL!=extSearch)
		{
			extSearch->VolumeAdded(newVlHd);
		}

		return newVlHd;
	}
	return NULL;
}

YSRESULT YsShellExt::ModifyVolume(VolumeHandle vlHd,YSSIZE_T nVlFg,const FaceGroupHandle vlFgHd[])
{
	Volume *vlProp=volumeArray[vlHd];
	if(NULL!=vlProp && YSTRUE!=volumeArray.IsFrozen(vlHd))
	{
		if(NULL!=extSearch)
		{
			extSearch->VolumeAdded(vlHd);
		}

		for(YSSIZE_T idx=0; idx<vlProp->fgHdArray.GetN(); ++idx)
		{
			FaceGroup *fgProp=faceGroupArray[vlProp->fgHdArray[idx]];
			--fgProp->refCount;
		}

		vlProp->fgHdArray.Set(nVlFg,vlFgHd);

		for(YSSIZE_T idx=0; idx<vlProp->fgHdArray.GetN(); ++idx)
		{
			FaceGroup *fgProp=faceGroupArray[vlProp->fgHdArray[idx]];
			++fgProp->refCount;
		}

		if(NULL!=extSearch)
		{
			extSearch->VolumeAdded(vlHd);
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::AddVolumeFaceGroup(VolumeHandle vlHd,YSSIZE_T nVlFg,const FaceGroupHandle vlFgHd[])
{
	Volume *vlProp=volumeArray[vlHd];
	if(NULL!=vlProp && YSTRUE!=volumeArray.IsFrozen(vlHd))
	{
		YSSIZE_T idx0=vlProp->fgHdArray.GetN();
		vlProp->fgHdArray.Append(nVlFg,vlFgHd);
		for(YSSIZE_T idx=idx0; idx<vlProp->fgHdArray.GetN(); ++idx)
		{
			FaceGroup *fgProp=faceGroupArray[vlProp->fgHdArray[idx]];
			++fgProp->refCount;
		}

		if(NULL!=extSearch)
		{
			extSearch->MultipleFaceGroupAddedToVolume(vlHd,nVlFg,vlFgHd);
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::FreezeVolume(YsShellExt::VolumeHandle vlHd)
{
	if(NULL!=vlHd && YSTRUE!=volumeArray.IsFrozen(vlHd))
	{
		if(NULL!=extSearch)
		{
			extSearch->VolumeAboutToBeDeleted(vlHd);
		}

		Volume *vlProp=volumeArray[vlHd];

		for(YSSIZE_T idx=0; idx<vlProp->fgHdArray.GetN(); ++idx)
		{
			FaceGroup *fgProp=faceGroupArray[vlProp->fgHdArray[idx]];
			--fgProp->refCount;
			++fgProp->refCountFromFrozenEntity;
		}
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::MeltVolume(YsShellExt::VolumeHandle vlHd)
{
	if(NULL!=vlHd && YSTRUE==volumeArray.IsFrozen(vlHd))
	{
		volumeArray.Melt(vlHd);

		Volume *vlProp=volumeArray[vlHd];
		for(YSSIZE_T idx=0; idx<vlProp->fgHdArray.GetN(); ++idx)
		{
			FaceGroup *fgProp=faceGroupArray[vlProp->fgHdArray[idx]];
			++fgProp->refCount;
			--fgProp->refCountFromFrozenEntity;
		}

		if(NULL!=extSearch)
		{
			extSearch->VolumeAdded(vlHd);
		}

		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteVolume(YsShellExt::VolumeHandle vlHd)
{
	Volume *vlProp=volumeArray[vlHd];
	if(NULL!=vlProp && YSTRUE!=volumeArray.IsFrozen(vlHd))
	{
		if(NULL!=extSearch)
		{
			extSearch->VolumeAboutToBeDeleted(vlHd);
		}

		for(YSSIZE_T idx=0; idx<vlProp->fgHdArray.GetN(); ++idx)
		{
			FaceGroup *fgProp=faceGroupArray[vlProp->fgHdArray[idx]];
			--fgProp->refCount;
		}
		volumeArray.Delete(vlHd);
		return YSOK;
	}
	return YSERR;
}

YSRESULT YsShellExt::DeleteFrozenVolume(YsShellExt::VolumeHandle vlHd)
{
	if(YSTRUE==volumeArray.IsFrozen(vlHd))
	{
		Volume *vlProp=volumeArray.GetObjectMaybeFrozen(vlHd);
		if(NULL!=vlProp)
		{
			for(YSSIZE_T idx=0; idx<vlProp->fgHdArray.GetN(); ++idx)
			{
				FaceGroup *fgProp=faceGroupArray.GetObjectMaybeFrozen(vlProp->fgHdArray[idx]);
				--fgProp->refCountFromFrozenEntity;
			}
			volumeArray.DeleteFrozen(vlHd);
			return YSOK;
		}
	}
	return YSERR;
}

#include "ysshellextsearch.h"

YSRESULT YsShellExt::SetVolumeLabel(VolumeHandle vlHd,const char label[])
{
	Volume *vlProp=volumeArray[vlHd];
	if(NULL!=vlProp && YSTRUE!=volumeArray.IsFrozen(vlHd))
	{
		vlProp->label=label;
		return YSOK;
	}
	return YSERR;
}

YSSIZE_T YsShellExt::GetNumEdge(void) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->GetNumEdge(*this);
	}
	return 0;
}

YSRESULT YsShellExt::MoveToNextEdge(YsShellEdgeEnumHandle &handle) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->FindNextEdge(*this,handle);
	}
	return YSERR;
}

YSRESULT YsShellExt::GetEdge(YsShellVertexHandle &edVtHd1,YsShellVertexHandle &edVtHd2,YsShellEdgeEnumHandle handle) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->GetEdge(*this,edVtHd1,edVtHd2,handle);
	}
	edVtHd1=NULL;
	edVtHd2=NULL;
	return YSERR;
}

YSRESULT YsShellExt::GetEdge(YsShellVertexHandle edVtHd[2],YsShellEdgeEnumHandle edHd) const
{
	return GetEdge(edVtHd[0],edVtHd[1],edHd);
}

YSRESULT YsShellExt::EnableSearch(void)
{
	if(NULL==shlSearch && NULL==extSearch)
	{
		shlSearch=new YsShellSearchTable;
		AttachSearchTable(shlSearch);
		extSearch=new SearchTable;
		extSearch->Attach(this);
		return YSOK;
	}
	return YSERR;
}

YSBOOL YsShellExt::IsSearchEnabled(void) const
{
	if(NULL!=shlSearch && NULL!=extSearch)
	{
		return YSTRUE;
	}
	return YSFALSE;
}

YSRESULT YsShellExt::DisableSearch(void)
{
	if(NULL!=shlSearch)
	{
		YsShell::DetachSearchTable();
		delete shlSearch;
		shlSearch=NULL;
	}
	if(NULL!=extSearch)
	{
		extSearch->Detach();
		delete extSearch;
		extSearch=NULL;
	}
	return YSOK;
}

YSRESULT YsShellExt::TestSearchTable(void) const
{
	YSRESULT res=YSOK;
	if(NULL!=extSearch)
	{
		// Const-edge to vertex
		ConstEdgeHandle ceHd=NULL;
		while(YSOK==MoveToNextConstEdge(ceHd))
		{
			YSSIZE_T nVt;
			const YsShellVertexHandle *ceVtHd;
			YSBOOL isLoop;

			if(YSOK==GetConstEdge(nVt,ceVtHd,isLoop,ceHd))
			{
				for(YSSIZE_T idx=0; idx<nVt; ++idx)
				{
					YsArray <ConstEdgeHandle,16> vtCeHd;
					if(YSOK==extSearch->FindConstEdgeFromVertex <16> (vtCeHd,ceVtHd[idx]))
					{
						if(YSTRUE!=vtCeHd.IsIncluded(ceHd))
						{
							printf("%s %d\n",__FUNCTION__,__LINE__);
							printf("Error!  Const-edge to vertex table does not include what's supposed to be included.\n");
							res=YSERR;
						}
					}
					else
					{
						printf("%s %d\n",__FUNCTION__,__LINE__);
						printf("Error!  Const-edge cannot be found from a vertex.\n");
						res=YSERR;
					}
				}
			}
		}

		// Vertex to const edge
		YsShellVertexHandle vtHd=NULL;
		while(YSOK==MoveToNextVertex(vtHd))
		{
			YsArray <ConstEdgeHandle,16> vtCeHd;
			if(YSOK==extSearch->FindConstEdgeFromVertex(vtCeHd,vtHd))
			{
				for(YSSIZE_T idx=0; idx<vtCeHd.GetN(); ++idx)
				{
					YSSIZE_T nVt;
					const YsShellVertexHandle *ceVtHd;
					YSBOOL isLoop;

					if(YSOK==GetConstEdge(nVt,ceVtHd,isLoop,vtCeHd[idx]))
					{
						if(YSTRUE!=YsIsIncluded(nVt,ceVtHd,vtHd))
						{
							printf("%s %d\n",__FUNCTION__,__LINE__);
							printf("Error!  Const-edge is found from a vertex, but the const-edge is not using the vertex.\n");
							res=YSERR;
						}
					}
					else
					{
						printf("%s %d\n",__FUNCTION__,__LINE__);
						printf("Error!  Const-edge is found from a vertex, but the const-edge does not exist.\n");
						res=YSERR;
					}
				}
			}
		}

		// FaceGroup to Polygon
		for(FaceGroupHandle fgHd=NULL; YSOK==MoveToNextFaceGroup(fgHd); )
		{
			YSSIZE_T nFgPl;
			const YsShellPolygonHandle *fgPlHd;
			if(YSOK==GetFaceGroup(nFgPl,fgPlHd,fgHd))
			{
				for(YSSIZE_T idx=0; idx<nFgPl; ++idx)
				{
					FaceGroupHandle found=extSearch->FindFaceGroupFromPolygon(fgPlHd[idx]);
					if(found!=fgHd)
					{
						printf("%s %d\n",__FUNCTION__,__LINE__);
						printf("Error!  Polygon to face group table is corrupted.\n");
						res=YSERR;
					}
				}
			}
			else
			{
				printf("%s %d\n",__FUNCTION__,__LINE__);
				printf("Error!  Cannot get face group polygons.\n");
				res=YSERR;
			}
		}

		// Polygon to FaceGroup
		for(YsShellPolygonHandle plHd=NULL; YSOK==MoveToNextPolygon(plHd); )
		{
			FaceGroupHandle fgHd=extSearch->FindFaceGroupFromPolygon(plHd);
			if(NULL!=fgHd)
			{
				YSSIZE_T nFgPl;
				const YsShellPolygonHandle *fgPlHd;
				GetFaceGroup(nFgPl,fgPlHd,fgHd);
				if(YSTRUE!=YsIsIncluded(nFgPl,fgPlHd,plHd))
				{
					printf("%s %d\n",__FUNCTION__,__LINE__);
					printf("Error!  Polygon to FaceGroup table is corrupted.\n");
					res=YSERR;
				}
			}
		}

		// Volume to FaceGroup
		for(VolumeHandle vlHd=NULL; YSOK==MoveToNextVolume(vlHd); )
		{
			YSSIZE_T nVlFg;
			const YsShellExt::FaceGroupHandle *vlFgHd;
			if(YSOK==GetVolume(nVlFg,vlFgHd,vlHd))
			{
				for(YSSIZE_T idx=0; idx<nVlFg; ++idx)
				{
					YSSIZE_T nFound=0;
					const VolumeHandle *found=NULL;
					if(YSOK!=extSearch->FindVolumeFromFaceGroup(nFound,found,vlFgHd[idx]) ||
					   YSTRUE!=YsIsIncluded(nFound,found,vlHd))
					{
						printf("%s %d\n",__FUNCTION__,__LINE__);
						printf("Error!  Face group to volume table is corrupted.\n");
						res=YSERR;
					}
				}
			}
			else
			{
				printf("%s %d\n",__FUNCTION__,__LINE__);
				printf("Error!  Cannot get volume face groups.\n");
				res=YSERR;
			}
		}

		// FaceGroup to Volume
		for(YsShellExt::FaceGroupHandle fgHd=NULL; YSOK==MoveToNextFaceGroup(fgHd); )
		{
			YSSIZE_T nVl;
			const YsShellExt::VolumeHandle *vlHd;

			if(YSOK==extSearch->FindVolumeFromFaceGroup(nVl,vlHd,fgHd))
			{
				for(YSSIZE_T vlIdx=0; vlIdx<nVl; ++vlIdx)
				{
					YSSIZE_T nVlFg;
					const YsShellExt::FaceGroupHandle *vlFgHd;
					GetVolume(nVlFg,vlFgHd,vlHd[vlIdx]);
					if(YSTRUE!=YsIsIncluded(nVlFg,vlFgHd,fgHd))
					{
						printf("%s %d\n",__FUNCTION__,__LINE__);
						printf("Error!  FaceGroup to Volume table is corrupted.\n");
						res=YSERR;
					}
				}
			}
		}
	}

	return res;
}

YsShellVertexHandle YsShellExt::FindVertex(YSHASHKEY vtKey) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->FindVertex(*this,vtKey);
	}
	return NULL;
}

YsShellPolygonHandle YsShellExt::FindPolygon(YSHASHKEY plKey) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->FindPolygon(*this,plKey);
	}
	return NULL;
}

int YsShellExt::GetNumPolygonUsingEdge(YsShellEdgeEnumHandle edHd) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->GetNumPolygonUsingEdge(*this,edHd);
	}
	return 0;
}

int YsShellExt::GetNumPolygonUsingEdge(const YsShellVertexHandle edVtHd[2]) const
{
	return GetNumPolygonUsingEdge(edVtHd[0],edVtHd[1]);
}

int YsShellExt::GetNumPolygonUsingEdge(YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->GetNumPolygonUsingEdge(*this,edVtHd0,edVtHd1);
	}
	return 0;
}

int YsShellExt::GetNumPolygonUsingVertex(YsShellVertexHandle vtHd) const
{
	if(NULL!=shlSearch)
	{
		int nVtPl;
		const YsShellPolygonHandle *plHdArray;
		if(YSOK==shlSearch->FindPolygonListByVertex(nVtPl,plHdArray,*this,vtHd))
		{
			return nVtPl;
		}
	}
	return 0;
}

YSRESULT YsShellExt::FindPolygonFromVertex(YSSIZE_T &nPl,const YsShellPolygonHandle *&plHdArray,YsShellVertexHandle vtHd) const
{
	if(NULL!=shlSearch)
	{
		int nVtPl;
		if(YSOK==shlSearch->FindPolygonListByVertex(nVtPl,plHdArray,*this,vtHd))
		{
			nPl=nVtPl;
			return YSOK;
		}
	}
	return YSERR;
}

YsArray <YsShellPolygonHandle> YsShellExt::FindPolygonFromVertex(YsShellVertexHandle vtHd) const
{
	YsArray <YsShellPolygonHandle> vtPlHd;
	FindPolygonFromVertex(vtPlHd,vtHd);
	return vtPlHd;
}

YSRESULT YsShellExt::FindPolygonFromEdgePiece(YSSIZE_T &nPl,const YsShellPolygonHandle *&plHdArray,YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1) const
{
	if(NULL!=shlSearch)
	{
		int nEdPl;
		if(YSOK==shlSearch->FindPolygonListByEdge(nEdPl,plHdArray,*this,edVtHd0,edVtHd1))
		{
			nPl=nEdPl;
			return YSOK;
		}
	}
	return YSERR;
}

YSRESULT YsShellExt::FindPolygonFromEdgePiece(YSSIZE_T &nPl,const YsShellPolygonHandle *&plHdArray,const YsShellVertexHandle edVtHd[2]) const
{
	return FindPolygonFromEdgePiece(nPl,plHdArray,edVtHd[0],edVtHd[1]);
}

YsArray <YsShellPolygonHandle,2> YsShellExt::FindPolygonFromEdgePiece(const YsShellVertexHandle edVtHd[2]) const
{
	YsArray <YsShellPolygonHandle,2> edPlHd;
	FindPolygonFromEdgePiece(edPlHd,edVtHd);
	return edPlHd;
}

YsArray <YsShellPolygonHandle,2> YsShellExt::FindPolygonFromEdgePiece(const YsShellVertexHandle edVtHd0,const YsShellVertexHandle edVtHd1) const
{
	YsArray <YsShellPolygonHandle,2> edPlHd;
	FindPolygonFromEdgePiece(edPlHd,edVtHd0,edVtHd1);
	return edPlHd;
}

YsShellPolygonHandle YsShellExt::GetNeighborPolygon(YsShellPolygonHandle plHd,int n) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->GetNeighborPolygon(*this,plHd,n);
	}
	return NULL;
}

YsShellPolygonHandle YsShellExt::GetNeighborPolygon(YsShellPolygonHandle plHd,YsShellVertexHandle edVtHd[2]) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->GetNeighborPolygon(*this,plHd,edVtHd[0],edVtHd[1]);
	}
	return NULL;
}

YsShellPolygonHandle YsShellExt::GetNeighborPolygon(YsShellPolygonHandle plHd,YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1) const
{
	if(NULL!=shlSearch)
	{
		return shlSearch->GetNeighborPolygon(*this,plHd,edVtHd0,edVtHd1);
	}
	return NULL;
}

YsArray <YsShellVertexHandle> YsShellExt::GetConnectedVertex(YsShellVertexHandle fromVtHd)
{
	YsArray <YsShellVertexHandle> connVtHd;
	GetConnectedVertex(connVtHd,fromVtHd);
	return connVtHd;
}

YsArray <YsShellVertexHandle> YsShellExt::GetConstEdgeConnectedVertex(YsShellVertexHandle fromVtHd) const
{
	YsArray <YsShellVertexHandle> connVtHd;
	GetConstEdgeConnectedVertex(connVtHd,fromVtHd);
	return connVtHd;
}

YsShellExt::ConstEdgeHandle YsShellExt::FindConstEdge(YSHASHKEY ceKey) const
{
	return constEdgeArray.FindObject(ceKey);
}

YSRESULT YsShellExt::FindConstEdgeFromVertex(YSSIZE_T &nCe,const YsShellExt::ConstEdgeHandle *&ceHdPtr,YsShellVertexHandle vtHd) const
{
	if(NULL!=extSearch)
	{
		return extSearch->FindConstEdgeFromVertex(nCe,ceHdPtr,vtHd);
	}
	return YSERR;
}

YsArray <YsShellExt::ConstEdgeHandle> YsShellExt::FindConstEdgeFromVertex(YsShellVertexHandle vtHd) const
{
	YsArray <YsShellExt::ConstEdgeHandle> vtCeHd;
	FindConstEdgeFromVertex(vtCeHd,vtHd);
	return vtCeHd;
}

int YsShellExt::GetNumConstEdgeUsingVertex(YsShellVertexHandle vtHd) const
{
	if(NULL!=extSearch)
	{
		YSSIZE_T nCe;
		const ConstEdgeHandle *ceHdPtr;
		if(YSOK==extSearch->FindConstEdgeFromVertex(nCe,ceHdPtr,vtHd))
		{
			return (int)nCe;
		}
	}
	return 0;
}

int YsShellExt::GetNumConstEdgeUsingEdgePiece(YsShellVertexHandle edVtHd0,YsShellVertexHandle edVtHd1) const
{
	YsArray <ConstEdgeHandle,16> ceHdArray;
	if(YSOK==FindConstEdgeFromEdgePiece(ceHdArray,edVtHd0,edVtHd1))
	{
		return (int)ceHdArray.GetN();
	}
	return 0;
}

int YsShellExt::GetNumConstEdgeUsingEdgePiece(YsShellVertexHandle edVtHd[2]) const
{
	return GetNumConstEdgeUsingEdgePiece(edVtHd[0],edVtHd[1]);
}

YsShellExt::FaceGroupHandle YsShellExt::FindFaceGroup(YSHASHKEY fgKey) const
{
	return faceGroupArray.FindObject(fgKey);
}

YsShellExt::FaceGroupHandle YsShellExt::FindFaceGroupFromPolygon(YsShellPolygonHandle plHd) const
{
	if(NULL!=plHd && NULL!=extSearch)
	{
		return extSearch->FindFaceGroupFromPolygon(plHd);
	}
	return NULL;
}

YsArray <YsShellExt::FaceGroupHandle> YsShellExt::GetNeighborFaceGroup(YsShellExt::FaceGroupHandle fromFgHd)
{
	auto fgPlHdArray=GetFaceGroup(fromFgHd);

	YsShellPolygonStore seedPlHd(this->Conv());
	seedPlHd.AddPolygon(fgPlHdArray);

	FaceGroupStore neiFgHdStore(*this);

	for(auto fgPlHd : fgPlHdArray)
	{
		auto plVtHd=GetPolygonVertex(fgPlHd);
		for(int edIdx=0; edIdx<plVtHd.GetN(); ++edIdx)
		{
			const YsShellVertexHandle edVtHd[2]=
			{
				plVtHd[edIdx],
				plVtHd.GetCyclic(edIdx+1)
			};

			auto neiPlHdArray=FindPolygonFromEdgePiece(edVtHd);
			for(auto neiPlHd : neiPlHdArray)
			{
				if(YSTRUE!=seedPlHd.IsIncluded(neiPlHd))
				{
					auto neiFgHd=FindFaceGroupFromPolygon(neiPlHd);
					if(NULL!=neiFgHd && neiFgHd!=fromFgHd)
					{
						neiFgHdStore.AddFaceGroup(neiFgHd);
					}
				}
			}
		}
	}

	YsArray <FaceGroupHandle> neiFgHdArray;
	for(auto fgHd : neiFgHdStore)
	{
		neiFgHdArray.Append(fgHd);
	}

	return neiFgHdArray;
}

YsShellExt::VolumeHandle YsShellExt::FindVolume(YSHASHKEY fgKey) const
{
	return volumeArray.FindObject(fgKey);
}

