/* ////////////////////////////////////////////////////////////

  File Name: ysshellextdrawbuffer.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "ysshellext.h"
#include "ysshellextdrawbuffer.h"
#include <ystask.h>
#include <thread>

#define YS_RUN_PARALLEL

YsShellExtDrawingBuffer::YsShellExtDrawingBuffer()
{
	viewPortEnabled=YSFALSE;
	viewPort[0]=-YsXYZ();
	viewPort[1]=YsXYZ();

	drawBackFaceInDifferentColor=YSTRUE;
	backFaceColor.SetIntRGB(128,16,16);
}

void YsShellExtDrawingBuffer::EnableViewPort(void)
{
	viewPortEnabled=YSTRUE;
}

void YsShellExtDrawingBuffer::DisableViewPort(void)
{
	viewPortEnabled=YSFALSE;
}

void YsShellExtDrawingBuffer::SetViewPort(const YsVec3 viewPort[2])
{
	this->viewPort[0]=viewPort[0];
	this->viewPort[1]=viewPort[1];
}

void YsShellExtDrawingBuffer::RemakeNormalVertexBuffer(const class YsShellExt &shl,const double normalVertexMarkerRadius,const double roundVertexMarkerRadius)
{
	normalVertexPosBuffer.CleanUp();
	normalVertexPixOffsetBuffer.CleanUp();
	normalVertexIdxBuffer.CleanUp();

	roundVertexPosBuffer.CleanUp();
	roundVertexPixOffsetBuffer.CleanUp();
	roundVertexIdxBuffer.CleanUp();

	YsShellVertexHandle vtHd=NULL;
	while(YSOK==shl.MoveToNextVertex(vtHd))
	{
		YsVec3 pos;
		shl.GetVertexPosition(pos,vtHd);

		if(YSTRUE==viewPortEnabled && YSTRUE!=YsCheckInsideBoundingBox3(pos,viewPort[0],viewPort[1]))
		{
			continue;
		}

		const YsShellExt::VertexAttrib *vtAttr=shl.GetVertexAttrib(vtHd);
		if(YSTRUE!=vtAttr->IsRound())
		{
			normalVertexPosBuffer.AddVertex(pos);
			normalVertexPixOffsetBuffer.AddPixOffset(-normalVertexMarkerRadius,-normalVertexMarkerRadius);

			normalVertexPosBuffer.AddVertex(pos);
			normalVertexPixOffsetBuffer.AddPixOffset( normalVertexMarkerRadius*3.0f,-normalVertexMarkerRadius);

			normalVertexPosBuffer.AddVertex(pos);
			normalVertexPixOffsetBuffer.AddPixOffset(-normalVertexMarkerRadius,normalVertexMarkerRadius*3.0f);

			normalVertexIdxBuffer.Append((int)shl.GetSearchKey(vtHd));
		}
		else
		{
			roundVertexPosBuffer.AddVertex(pos);
			roundVertexPixOffsetBuffer.AddPixOffset(-roundVertexMarkerRadius,-roundVertexMarkerRadius);

			roundVertexPosBuffer.AddVertex(pos);
			roundVertexPixOffsetBuffer.AddPixOffset( roundVertexMarkerRadius*3.0f,-roundVertexMarkerRadius);

			roundVertexPosBuffer.AddVertex(pos);
			roundVertexPixOffsetBuffer.AddPixOffset(-roundVertexMarkerRadius,roundVertexMarkerRadius*3.0f);

			roundVertexIdxBuffer.Append((int)shl.GetSearchKey(vtHd));
		}
	}
}

void YsShellExtDrawingBuffer::RemakeSelectedVertexBuffer(const class YsShellExt &shl,const double selectedVertexMarkerRadius,YSSIZE_T nSel,const YsShellVertexHandle selVtHd[])
{
	selectedVertexPosBuffer.CleanUp();
	selectedVertexIdxBuffer.CleanUp();
	selectedVertexPixOffsetBuffer.CleanUp();
	selectedVertexColBuffer.CleanUp();

	selectedVertexLineBuffer.CleanUp();
	selectedVertexLineColBuffer.CleanUp();

	for(int idx=0; idx<nSel; ++idx)
	{
		YsVec3 pos;
		shl.GetVertexPosition(pos,selVtHd[idx]);

		if(YSTRUE==viewPortEnabled && YSTRUE!=YsCheckInsideBoundingBox3(pos,viewPort[0],viewPort[1]))
		{
			continue;
		}

		YsColor col=ColorGradation(idx,nSel);
		selectedVertexColBuffer.AddColor(col);

		selectedVertexPosBuffer.AddVertex(pos);
		selectedVertexPixOffsetBuffer.AddPixOffset(-selectedVertexMarkerRadius,-selectedVertexMarkerRadius);

		selectedVertexPosBuffer.AddVertex(pos);
		selectedVertexPixOffsetBuffer.AddPixOffset(selectedVertexMarkerRadius*3.0f,-selectedVertexMarkerRadius);

		selectedVertexPosBuffer.AddVertex(pos);
		selectedVertexPixOffsetBuffer.AddPixOffset(-selectedVertexMarkerRadius,selectedVertexMarkerRadius*3.0f);

		selectedVertexIdxBuffer.Append(shl.GetSearchKey(selVtHd[idx]));
	}

	for(int idx=0; idx<nSel-1; ++idx)
	{
		YsVec3 edVtPos[2];
		shl.GetVertexPosition(edVtPos[0],selVtHd[idx]);
		shl.GetVertexPosition(edVtPos[1],selVtHd[idx+1]);

		const YsColor col[2]={ColorGradation(idx,nSel),ColorGradation(idx+1,nSel)};

		if(YSTRUE!=viewPortEnabled)
		{
			selectedVertexLineBuffer.AddVertex(edVtPos[0]);
			selectedVertexLineBuffer.AddVertex(edVtPos[1]);
			selectedVertexLineColBuffer.AddColor(col[0]);
			selectedVertexLineColBuffer.AddColor(col[1]);
		}
		else
		{
			YsVec3 clippedEdVtPos[2];
			if(YSOK==YsClipLineSeg3(
			    clippedEdVtPos[0],clippedEdVtPos[1],edVtPos[0],edVtPos[1],viewPort[0],viewPort[1]))
			{
				selectedVertexLineBuffer.AddVertex(clippedEdVtPos[0]);
				selectedVertexLineBuffer.AddVertex(clippedEdVtPos[1]);
				selectedVertexLineColBuffer.AddColor(col[0]);
				selectedVertexLineColBuffer.AddColor(col[1]);
			}
		}
	}
}

////////////////////////////////////////////////////////////

class YsShellExtDrawingBuffer::RemakePolygonBufferTask : public YsTask
{
public:
	// Output
	YsGLVertexBuffer normalEdgePosBuffer;
	YsArray <int> normalEdgeIdxBuffer;
	YsGLVertexBuffer shrunkEdgePosBuffer;
	YsArray <int> shrunkEdgeIdxBuffer;

	YsGLVertexBuffer solidShadedVtxBuffer;
	YsGLNormalBuffer solidShadedNomBuffer;
	YsGLColorBuffer solidShadedColBuffer;
	YsArray <int> solidShadedIdxBuffer;

	YsGLVertexBuffer solidUnshadedVtxBuffer;
	YsGLColorBuffer solidUnshadedColBuffer;
	YsArray <int> solidUnshadedIdxBuffer;

	YsGLVertexBuffer trspShadedVtxBuffer;
	YsGLNormalBuffer trspShadedNomBuffer;
	YsGLColorBuffer trspShadedColBuffer;
	YsArray <int> trspShadedIdxBuffer;

	YsGLVertexBuffer trspUnshadedVtxBuffer;
	YsGLColorBuffer trspUnshadedColBuffer;
	YsArray <int> trspUnshadedIdxBuffer;

	YsGLVertexBuffer backFaceVtxBuffer;
	YsGLColorBuffer backFaceColBuffer;
	YsArray <int> backFaceIdxBuffer;

	// Input
	YsArray <YsShellVertexHandle> edVtHdArray;
	YsArray <YsShellPolygonHandle> plHdArray;
	const YsShellExt *shlExtPtr;
	double polygonShrinkRatioForPicking;
	YSBOOL viewPortEnabled;
	YsVec3 viewPort[2];
	YSBOOL drawBackFaceInDifferentColor;
	YsColor backFaceColor;
	const YsHashTable <YsVec3> *vtxNomHashPtr;

	virtual void Run(void);  // Implementation of the thread task.

	void AddPolygonFragment(
	    const YsShellExt &shl,
	    YsShellPolygonHandle plHd,const YsVec3 &defNom,const YsHashTable <YsVec3> &vtxNomHash,const YsShellVertexHandle triVtHd[3],YSBOOL flatShaded,
	    YSBOOL addAsBackFace);
	void AddPolygonFragment(
	    const YsVec3 plVtPos[3],const YsVec3 plVtNom[3],unsigned int index,const YsColor &col,YSBOOL lighting,
	    YSBOOL addAsBackFace);
};

void YsShellExtDrawingBuffer::RemakePolygonBufferTask::Run(void)
{
	const YsShellExt &shl=*shlExtPtr;

	for(YSSIZE_T edIdx=0; edIdx<=edVtHdArray.GetN()-2; edIdx+=2)
	{
		const YsShellVertexHandle edVtHd[2]={edVtHdArray[edIdx],edVtHdArray[edIdx+1]};

		YsVec3 edVtPos[2];
		shl.GetVertexPosition(edVtPos[0],edVtHd[0]);
		shl.GetVertexPosition(edVtPos[1],edVtHd[1]);

		if(YSTRUE==viewPortEnabled)
		{
			YsVec3 clippedEdVtPos[2];
			if(YSOK==YsClipLineSeg3(
			    clippedEdVtPos[0],clippedEdVtPos[1],edVtPos[0],edVtPos[1],viewPort[0],viewPort[1]))
			{
				normalEdgePosBuffer.AddVertex(clippedEdVtPos[0]);
				normalEdgeIdxBuffer.Append(shl.GetSearchKey(edVtHd[0]));
				normalEdgePosBuffer.AddVertex(clippedEdVtPos[1]);
				normalEdgeIdxBuffer.Append(shl.GetSearchKey(edVtHd[1]));
			}
		}
		else
		{
			normalEdgePosBuffer.AddVertex(edVtPos[0]);
			normalEdgeIdxBuffer.Append(shl.GetSearchKey(edVtHd[0]));
			normalEdgePosBuffer.AddVertex(edVtPos[1]);
			normalEdgeIdxBuffer.Append(shl.GetSearchKey(edVtHd[1]));
		}
	}

	for(YSSIZE_T plIdx=0; plIdx<plHdArray.GetN(); ++plIdx)
	{
		YsShellPolygonHandle plHd=plHdArray[plIdx];
		YsArray <YsVec3,4> plVtPos;
		shl.GetVertexListOfPolygon(plVtPos,plHd);

		YsVec3 cen;
		shl.GetCenterOfPolygon(cen,plHd);

		for(YSSIZE_T idx=0; idx<plVtPos.GetN(); ++idx)
		{
			plVtPos[idx]=cen+(plVtPos[idx]-cen)*polygonShrinkRatioForPicking;
		}

		for(YSSIZE_T idx=0; idx<plVtPos.GetN(); ++idx)
		{
			if(YSTRUE==viewPortEnabled)
			{
				YsVec3 clippedEdVtPos[2];
				if(YSOK==YsClipLineSeg3(
				    clippedEdVtPos[0],clippedEdVtPos[1],plVtPos[idx],plVtPos.GetCyclic(idx+1),viewPort[0],viewPort[1]))
				{
					shrunkEdgePosBuffer.AddVertex(clippedEdVtPos[0]);
					shrunkEdgePosBuffer.AddVertex(clippedEdVtPos[1]);
					shrunkEdgeIdxBuffer.Append(shl.GetSearchKey(plHd));
				}
			}
			else
			{
				shrunkEdgePosBuffer.AddVertex(plVtPos[idx]);
				shrunkEdgePosBuffer.AddVertex(plVtPos.GetCyclic(idx+1));
				shrunkEdgeIdxBuffer.Append(shl.GetSearchKey(plHd));
			}
		}

		YsVec3 nom;
		shl.GetNormalOfPolygon(nom,plHd);

		const YSBOOL twoSide=(YsOrigin()==nom ? YSTRUE : YSFALSE);
		const YSBOOL flatShaded=twoSide;
		if(YSTRUE==twoSide)
		{
			YsGetAverageNormalVector(nom,plVtPos.GetN(),plVtPos);
		}

		const YSBOOL backFaceSwitch=((YSTRUE!=twoSide && YSTRUE==drawBackFaceInDifferentColor) ? YSTRUE : YSFALSE);


		const YsHashTable <YsVec3> &vtxNomHash=*vtxNomHashPtr;

		const YsShellVertexHandle *const plVtHd=shl.GetVertexListOfPolygon(plHd);;
		if(3==plVtPos.GetN())
		{
			AddPolygonFragment(shl,plHd,nom,vtxNomHash,plVtHd,flatShaded,YSFALSE);

			const YsShellVertexHandle reverseTriVtHd[3]=
			{
				plVtHd[2],plVtHd[1],plVtHd[0]
			};
			AddPolygonFragment(shl,plHd,-nom,vtxNomHash,reverseTriVtHd,flatShaded,backFaceSwitch);
		}
		else if(YSTRUE==YsCheckConvex3(plVtPos.GetN(),plVtPos))
		{
			for(int i=1; i<plVtPos.GetN()-1; i++)
			{
				YsShellVertexHandle triVtHd[3]=
				{
					plVtHd[0],
					plVtHd[i],
					plVtHd[i+1]
				};
				AddPolygonFragment(shl,plHd,nom,vtxNomHash,triVtHd,flatShaded,YSFALSE);

				YsSwapSomething(triVtHd[0],triVtHd[1]);
				AddPolygonFragment(shl,plHd,-nom,vtxNomHash,triVtHd,flatShaded,backFaceSwitch);
			}
		}
		else
		{
			YsSword swd;

			YsArray <int,16> plVtKey;
			plVtKey.Set(plVtPos.GetN(),NULL);
			for(int i=0; i<plVtPos.GetN(); i++)
			{
				plVtKey[i]=shl.GetSearchKey(plVtHd[i]);
			}

			swd.SetInitialPolygon(plVtPos.GetN(),plVtPos,plVtKey);
			swd.Triangulate();
			for(int i=0; i<swd.GetNumPolygon(); i++)
			{
				const YsArray <int> *triVtKey=swd.GetVertexIdList(i);
				if(triVtKey->GetN()>=3)
				{
					YsShellVertexHandle triVtHd[3]=
					{
						shl.FindVertex((*triVtKey)[0]),
						shl.FindVertex((*triVtKey)[1]),
						shl.FindVertex((*triVtKey)[2])
					};
					if(NULL!=triVtHd[0] && NULL!=triVtHd[1] && NULL!=triVtHd[2])
					{
						AddPolygonFragment(shl,plHd,nom,vtxNomHash,triVtHd,flatShaded,YSFALSE);
						YsSwapSomething(triVtHd[0],triVtHd[1]);
						AddPolygonFragment(shl,plHd,-nom,vtxNomHash,triVtHd,flatShaded,backFaceSwitch);
					}
				}
			}
		}
	}
}

void YsShellExtDrawingBuffer::RemakePolygonBufferTask::AddPolygonFragment(
    const YsShellExt &shl,
    YsShellPolygonHandle plHd,const YsVec3 &defNom,const YsHashTable <YsVec3> &vtxNomHash,const YsShellVertexHandle triVtHd[3],YSBOOL flatShaded,
    YSBOOL addAsBackFace)
{
	const YsShellExt::PolygonAttrib *plAttr=shl.GetPolygonAttrib(plHd);

	YsColor col;
	shl.GetColorOfPolygon(col,plHd);

	const YSBOOL lighting=YsReverseBool(plAttr->GetNoShading());

	YsVec3 triVtPos[3];
	YsVec3 triNom[3];

	for(int i=0; i<3; i++)
	{
		const YsShellExt::VertexAttrib *vtAttr=shl.GetVertexAttrib(triVtHd[i]);

		shl.GetVertexPosition(triVtPos[i],triVtHd[i]);
		if(YSTRUE!=flatShaded && YSTRUE==vtAttr->IsRound() && YSOK==vtxNomHash.FindElement(triNom[i],shl.GetSearchKey(triVtHd[i])))
		{
		}
		else
		{
			triNom[i]=defNom;
		}
	}

	if(YSTRUE!=viewPortEnabled)
	{
		AddPolygonFragment(triVtPos,triNom,shl.GetSearchKey(plHd),col,lighting,addAsBackFace);
	}
	else
	{
		int newNp;
		YsVec3 newp[32]; // 15 should be good enough.  Just in case.
		if(YSOK==YsClipPolygon3(newNp,newp,32,3,triVtPos,viewPort[0],viewPort[1]) && 3<=newNp)
		{
			YsVec3 clpPos[3],clpNom[3];
			clpPos[0]=newp[0];
			clpNom[0]=defNom;
			for(int i=0; i<3; ++i)
			{
				if(triVtPos[i]==clpPos[0])
				{
					clpNom[0]=triNom[i];
					break;
				}
			}

			for(int i=1; i<newNp-1; ++i)
			{
				clpPos[1]=newp[i];
				clpNom[1]=defNom;
				clpPos[2]=newp[i+1];
				clpNom[2]=defNom;

				for(int j=1; j<3; ++j)
				{
					for(int k=0; k<3; ++k)
					{
						if(clpPos[j]==triVtPos[k])
						{
							clpNom[j]=triNom[k];
							break;
						}
					}
				}

				AddPolygonFragment(clpPos,clpNom,shl.GetSearchKey(plHd),col,lighting,addAsBackFace);
			}
		}
	}
}

void YsShellExtDrawingBuffer::RemakePolygonBufferTask::AddPolygonFragment(
    const YsVec3 plVtPos[3],const YsVec3 plVtNom[3],unsigned int index,const YsColor &col,YSBOOL lighting,
    YSBOOL addAsBackFace)
{
	for(int i=0; i<3; ++i)
	{
		if(YSTRUE==addAsBackFace)
		{
			backFaceVtxBuffer.AddVertex(plVtPos[i]);
			backFaceColBuffer.AddColor(backFaceColor);
		}
		else if(YSTRUE==lighting)
		{
			if(0.99f<=col.Af())
			{
				solidShadedVtxBuffer.AddVertex(plVtPos[i]);
				solidShadedNomBuffer.AddNormal(plVtNom[i]);
				solidShadedColBuffer.AddColor(col);
			}
			else
			{
				trspShadedVtxBuffer.AddVertex(plVtPos[i]);
				trspShadedNomBuffer.AddNormal(plVtNom[i]);
				trspShadedColBuffer.AddColor(col);
			}
		}
		else
		{
			if(0.99f<=col.Af())
			{
				solidUnshadedVtxBuffer.AddVertex(plVtPos[i]);
				solidUnshadedColBuffer.AddColor(col);
			}
			else
			{
				trspUnshadedVtxBuffer.AddVertex(plVtPos[i]);
				trspUnshadedColBuffer.AddColor(col);
			}
		}
	}

	if(YSTRUE==addAsBackFace)
	{
		backFaceIdxBuffer.Append(index);
	}
	else if(YSTRUE==lighting)
	{
		if(0.99f<=col.Af())
		{
			solidShadedIdxBuffer.Append(index);
		}
		else
		{
			trspShadedIdxBuffer.Append(index);
		}
	}
	else
	{
		if(0.99f<=col.Af())
		{
			solidUnshadedIdxBuffer.Append(index);
		}
		else
		{
			trspUnshadedIdxBuffer.Append(index);
		}
	}
}

void YsShellExtDrawingBuffer::RemakePolygonBuffer(const class YsShellExt &shl,const double polygonShrinkRatioForPicking)
{
	normalEdgePosBuffer.CleanUp();
	normalEdgeIdxBuffer.CleanUp();
	shrunkEdgePosBuffer.CleanUp();
	shrunkEdgeIdxBuffer.CleanUp();

	// See ysdnmgl2.0.cpp
	solidShadedVtxBuffer.CleanUp();
	solidShadedNomBuffer.CleanUp();
	solidShadedColBuffer.CleanUp();
	solidShadedIdxBuffer.CleanUp();

	solidUnshadedVtxBuffer.CleanUp();
	solidUnshadedColBuffer.CleanUp();
	solidUnshadedIdxBuffer.CleanUp();

	trspShadedVtxBuffer.CleanUp();
	trspShadedNomBuffer.CleanUp();
	trspShadedColBuffer.CleanUp();
	trspShadedIdxBuffer.CleanUp();

	trspUnshadedVtxBuffer.CleanUp();
	trspUnshadedColBuffer.CleanUp();
	trspUnshadedIdxBuffer.CleanUp();

	backFaceVtxBuffer.CleanUp();
	backFaceColBuffer.CleanUp();
	backFaceIdxBuffer.CleanUp();

	lightVtxBuffer.CleanUp();
	lightColBuffer.CleanUp();
	lightPixOffsetBuffer.CleanUp();

	int nThread=(int)std::thread::hardware_concurrency();
	if(0==nThread)
	{
		nThread=4;
	}
	YsArray <YsShellExtDrawingBuffer::RemakePolygonBufferTask,32> taskArray(nThread,NULL);
	YsArray <YsTask *,32> taskPtrArray(nThread,NULL);

	{
		int taskIdx=0;
		for(YsShellEdgeEnumHandle edHd=NULL; YSOK==shl.MoveToNextEdge(edHd); )
		{
			YsShellVertexHandle edVtHd[2];
			shl.GetEdge(edVtHd[0],edVtHd[1],edHd);

			taskArray[taskIdx].edVtHdArray.Append(2,edVtHd);
			++taskIdx;
			if(nThread<=taskIdx)
			{
				taskIdx=0;
			}
		}
	}

	YsHashTable <YsVec3> vtxNomHash;
	{
		MakeVertexNormalCache(vtxNomHash,shl);

		for(int i=0; i<nThread; ++i)
		{
			taskArray[i].shlExtPtr=&shl;
			taskArray[i].polygonShrinkRatioForPicking=polygonShrinkRatioForPicking;
			taskArray[i].viewPortEnabled=viewPortEnabled;
			taskArray[i].viewPort[0]=viewPort[0];
			taskArray[i].viewPort[1]=viewPort[1];
			taskArray[i].drawBackFaceInDifferentColor=drawBackFaceInDifferentColor;
			taskArray[i].backFaceColor=backFaceColor;
			taskArray[i].vtxNomHashPtr=&vtxNomHash;
			taskPtrArray[i]=&taskArray[i];
		}

		int idx=0;
		for(YsShellPolygonHandle plHd=NULL; YSOK==shl.MoveToNextPolygon(plHd); )
		{
			taskArray[idx].plHdArray.Append(plHd);
			++idx;
			if(nThread<=idx)
			{
				idx=0;
			}
		}
	}

#ifdef YS_RUN_PARALLEL
	YsTask::RunParallel(taskPtrArray);
#else
	printf("RemakePolygonBuffer is running in sequential mode.\n");
	for(int i=0; i<nThread; ++i) // Will launch in parallel soon.
	{
		task[i].StartLocal();
	}
#endif

	for(int i=0; i<nThread; ++i)
	{
		normalEdgePosBuffer.MoveFrom(taskArray[i].normalEdgePosBuffer);
		normalEdgeIdxBuffer.Append(taskArray[i].normalEdgeIdxBuffer);
		shrunkEdgePosBuffer.MoveFrom(taskArray[i].shrunkEdgePosBuffer);
		shrunkEdgeIdxBuffer.Append(taskArray[i].shrunkEdgeIdxBuffer);

		solidShadedVtxBuffer.MoveFrom(taskArray[i].solidShadedVtxBuffer);
		solidShadedNomBuffer.MoveFrom(taskArray[i].solidShadedNomBuffer);
		solidShadedColBuffer.MoveFrom(taskArray[i].solidShadedColBuffer);
		solidShadedIdxBuffer.Append(taskArray[i].solidShadedIdxBuffer);

		solidUnshadedVtxBuffer.MoveFrom(taskArray[i].solidUnshadedVtxBuffer);
		solidUnshadedColBuffer.MoveFrom(taskArray[i].solidUnshadedColBuffer);
		solidUnshadedIdxBuffer.Append(taskArray[i].solidUnshadedIdxBuffer);

		trspShadedVtxBuffer.MoveFrom(taskArray[i].trspShadedVtxBuffer);
		trspShadedNomBuffer.MoveFrom(taskArray[i].trspShadedNomBuffer);
		trspShadedColBuffer.MoveFrom(taskArray[i].trspShadedColBuffer);
		trspShadedIdxBuffer.Append(taskArray[i].trspShadedIdxBuffer);

		backFaceVtxBuffer.MoveFrom(taskArray[i].backFaceVtxBuffer);
		backFaceColBuffer.MoveFrom(taskArray[i].backFaceColBuffer);
		backFaceIdxBuffer.Append(taskArray[i].backFaceIdxBuffer);

		trspUnshadedVtxBuffer.MoveFrom(taskArray[i].trspUnshadedVtxBuffer);
		trspUnshadedColBuffer.MoveFrom(taskArray[i].trspUnshadedColBuffer);
		trspUnshadedIdxBuffer.Append(taskArray[i].trspUnshadedIdxBuffer);
	}
}

////////////////////////////////////////////////////////////

class YsShellExtDrawingBuffer::RemakeSelectedPolygonBufferTask : public YsTask
{
public:
	// Output
	YsGLVertexBuffer selectedPolygonPosBuffer;
	YsGLColorBuffer selectedPolygonColBuffer;
	YsArray <int> selectedPolygonIdxBuffer;

	// Input
	YsArray <YsShellPolygonHandle> selPlHd;
	const YsShellExt *shlExtPtr;
	double polygonShrinkRatio;
	YSBOOL viewPortEnabled;
	YsVec3 viewPort[2];

	virtual void Run(void);  // Implementation of the thread task.
};

void YsShellExtDrawingBuffer::RemakeSelectedPolygonBufferTask::Run(void)
{
	const YsShellExt &shl=*shlExtPtr;
	for(YSSIZE_T plIdx=0; plIdx<selPlHd.GetN(); ++plIdx)
	{
		YsArray <YsVec3,4> plVtPos;
		shl.GetVertexListOfPolygon(plVtPos,selPlHd[plIdx]);

		YsVec3 cen;
		shl.GetCenterOfPolygon(cen,selPlHd[plIdx]);

		for(YSSIZE_T vtIdx=0; vtIdx<plVtPos.GetN(); ++vtIdx)
		{
			plVtPos[vtIdx]=cen+(plVtPos[vtIdx]-cen)*polygonShrinkRatio;
		}

		for(YSSIZE_T vtIdx=0; vtIdx<plVtPos.GetN(); ++vtIdx)
		{
			if(YSTRUE==viewPortEnabled)
			{
				YsVec3 clp[2];
				if(YSOK==YsClipLineSeg3(clp[0],clp[1],plVtPos[vtIdx],plVtPos.GetCyclic(vtIdx+1),viewPort[0],viewPort[1]))
				{
					selectedPolygonPosBuffer.AddVertex(clp[0]);
					selectedPolygonColBuffer.AddColor(ColorGradation(vtIdx,plVtPos.GetN()+1));
					selectedPolygonPosBuffer.AddVertex(clp[1]);
					selectedPolygonColBuffer.AddColor(ColorGradation(vtIdx+1,plVtPos.GetN()+1));
					selectedPolygonIdxBuffer.Append(shl.GetSearchKey(selPlHd[plIdx]));
				}
			}
			else
			{
				selectedPolygonPosBuffer.AddVertex(plVtPos[vtIdx]);
				selectedPolygonColBuffer.AddColor(ColorGradation(vtIdx,plVtPos.GetN()+1));
				selectedPolygonPosBuffer.AddVertex(plVtPos.GetCyclic(vtIdx+1));
				selectedPolygonColBuffer.AddColor(ColorGradation(vtIdx+1,plVtPos.GetN()+1));
				selectedPolygonIdxBuffer.Append(shl.GetSearchKey(selPlHd[plIdx]));
			}
		}
	}
}

void YsShellExtDrawingBuffer::RemakeSelectedPolygonBuffer(
    const class YsShellExt &shl,const double polygonShrinkRatio,YSSIZE_T nSel,const YsShellPolygonHandle selPlHd[])
{
	selectedPolygonPosBuffer.CleanUp();
	selectedPolygonColBuffer.CleanUp();
	selectedPolygonIdxBuffer.CleanUp();

	int nThread=(int)std::thread::hardware_concurrency();
	if(0==nThread)
	{
		nThread=4;
	}
	YsArray <YsShellExtDrawingBuffer::RemakeSelectedPolygonBufferTask,32> taskArray(nThread,NULL);
	YsArray <YsTask *,32> taskPtrArray(nThread,NULL);

	for(int taskIdx=0; taskIdx<nThread; ++taskIdx)
	{
		taskArray[taskIdx].shlExtPtr=&shl;
		taskArray[taskIdx].polygonShrinkRatio=polygonShrinkRatio;
		taskArray[taskIdx].viewPortEnabled=viewPortEnabled;
		taskArray[taskIdx].viewPort[0]=viewPort[0];
		taskArray[taskIdx].viewPort[1]=viewPort[1];
		taskPtrArray[taskIdx]=&taskArray[taskIdx];
	}

	int taskIdx=0;
	for(YSSIZE_T plIdx=0; plIdx<nSel; ++plIdx)
	{
		taskArray[taskIdx].selPlHd.Append(selPlHd[plIdx]);
		++taskIdx;
		if(nThread<=taskIdx)
		{
			taskIdx=0;
		}
	}

#ifdef YS_RUN_PARALLEL
	YsTask::RunParallel(taskPtrArray);
#else
	for(int i=0; i<nThread; ++i) // Will launch in parallel soon.
	{
		taskArray[i].StartLocal();
	}
#endif

	for(YSSIZE_T taskIdx=0; taskIdx<nThread; ++taskIdx)
	{
		selectedPolygonPosBuffer.MoveFrom(taskArray[taskIdx].selectedPolygonPosBuffer);
		selectedPolygonColBuffer.MoveFrom(taskArray[taskIdx].selectedPolygonColBuffer);
		selectedPolygonIdxBuffer.Append(taskArray[taskIdx].selectedPolygonIdxBuffer);
	}
}

void YsShellExtDrawingBuffer::RemakeConstEdgeBuffer(const class YsShellExt &shl,YSSIZE_T nSel,const YsShellExt::ConstEdgeHandle ceHdArray[])
{
	YsKeyStore selCeKey;
	for(YSSIZE_T idx=0; idx<nSel; ++idx)
	{
		selCeKey.AddKey(shl.GetSearchKey(ceHdArray[idx]));
	}

	constEdgeVtxBuffer.CleanUp();
	constEdgeIdxBuffer.CleanUp();

	YsShellExt::ConstEdgeHandle ceHd=NULL;
	while(YSOK==shl.MoveToNextConstEdge(ceHd))
	{
		if(YSTRUE==selCeKey.IsIncluded(shl.GetSearchKey(ceHd)))
		{
			continue;
		}

		YSBOOL isLoop;
		YSSIZE_T nVt;
		const YsShellVertexHandle *ceVtHd;

		shl.GetConstEdge(nVt,ceVtHd,isLoop,ceHd);

		for(YSSIZE_T vtIdx=0; vtIdx<nVt-1; ++vtIdx)
		{
			YsVec3 edVtPos[2];

			shl.GetVertexPosition(edVtPos[0],ceVtHd[vtIdx]);
			shl.GetVertexPosition(edVtPos[1],ceVtHd[vtIdx+1]);

			if(YSTRUE!=viewPortEnabled)
			{
				constEdgeVtxBuffer.AddVertex(edVtPos[0]);
				constEdgeVtxBuffer.AddVertex(edVtPos[1]);

				constEdgeIdxBuffer.Append(shl.GetSearchKey(ceHd));
			}
			else
			{
				YsVec3 clippedEdVtPos[2];
				if(YSOK==YsClipLineSeg3(
				    clippedEdVtPos[0],clippedEdVtPos[1],edVtPos[0],edVtPos[1],viewPort[0],viewPort[1]))
				{
					constEdgeVtxBuffer.AddVertex(clippedEdVtPos[0]);
					constEdgeVtxBuffer.AddVertex(clippedEdVtPos[1]);

					constEdgeIdxBuffer.Append(shl.GetSearchKey(ceHd));
				}
			}
		}

		if(YSTRUE==isLoop)
		{
			YsVec3 edVtPos[2];

			shl.GetVertexPosition(edVtPos[0],ceVtHd[nVt-1]);
			shl.GetVertexPosition(edVtPos[1],ceVtHd[0]);

			if(YSTRUE!=viewPortEnabled)
			{
				constEdgeVtxBuffer.AddVertex(edVtPos[0]);
				constEdgeVtxBuffer.AddVertex(edVtPos[1]);

				constEdgeIdxBuffer.Append(shl.GetSearchKey(ceHd));
			}
			else
			{
				YsVec3 clippedEdVtPos[2];
				if(YSOK==YsClipLineSeg3(
				    clippedEdVtPos[0],clippedEdVtPos[1],edVtPos[0],edVtPos[1],viewPort[0],viewPort[1]))
				{
					constEdgeVtxBuffer.AddVertex(clippedEdVtPos[0]);
					constEdgeVtxBuffer.AddVertex(clippedEdVtPos[1]);

					constEdgeIdxBuffer.Append(shl.GetSearchKey(ceHd));
				}
			}
		}
	}
}

void YsShellExtDrawingBuffer::RemakeSelectedConstEdgeBuffer(const class YsShellExt &shl,YSSIZE_T nSel,const YsShellExt::ConstEdgeHandle ceHdArray[])
{
	selectedConstEdgeVtxBuffer.CleanUp();
	selectedConstEdgeIdxBuffer.CleanUp();
	selectedConstEdgeArrowHeadVtxBuffer.CleanUp();

	for(YSSIZE_T idx=0; idx<nSel; ++idx)
	{
		YSBOOL isLoop;
		YSSIZE_T nVt;
		const YsShellVertexHandle *ceVtHd;

		shl.GetConstEdge(nVt,ceVtHd,isLoop,ceHdArray[idx]);

		for(YSSIZE_T vtIdx=0; vtIdx<nVt-1; ++vtIdx)
		{
			YsVec3 edVtPos[2];

			shl.GetVertexPosition(edVtPos[0],ceVtHd[vtIdx]);
			shl.GetVertexPosition(edVtPos[1],ceVtHd[vtIdx+1]);

			if(YSTRUE!=viewPortEnabled)
			{
				selectedConstEdgeVtxBuffer.AddVertex(edVtPos[0]);
				selectedConstEdgeVtxBuffer.AddVertex(edVtPos[1]);

				selectedConstEdgeIdxBuffer.Append(shl.GetSearchKey(ceHdArray[idx]));

				selectedConstEdgeArrowHeadVtxBuffer.AddArrowHeadTriangle(edVtPos[0],edVtPos[1],0.2);
			}
			else
			{
				YsVec3 clippedEdVtPos[2];
				if(YSOK==YsClipLineSeg3(
				    clippedEdVtPos[0],clippedEdVtPos[1],edVtPos[0],edVtPos[1],viewPort[0],viewPort[1]))
				{
					selectedConstEdgeVtxBuffer.AddVertex(clippedEdVtPos[0]);
					selectedConstEdgeVtxBuffer.AddVertex(clippedEdVtPos[1]);

					selectedConstEdgeIdxBuffer.Append(shl.GetSearchKey(ceHdArray[idx]));

					selectedConstEdgeArrowHeadVtxBuffer.AddArrowHeadTriangle(clippedEdVtPos[0],clippedEdVtPos[1],0.2);
				}
			}
		}

		if(YSTRUE==isLoop)
		{
			YsVec3 edVtPos[2];

			shl.GetVertexPosition(edVtPos[0],ceVtHd[nVt-1]);
			shl.GetVertexPosition(edVtPos[1],ceVtHd[0]);

			if(YSTRUE!=viewPortEnabled)
			{
				selectedConstEdgeVtxBuffer.AddVertex(edVtPos[0]);
				selectedConstEdgeVtxBuffer.AddVertex(edVtPos[1]);

				selectedConstEdgeIdxBuffer.Append(shl.GetSearchKey(ceHdArray[idx]));

				selectedConstEdgeArrowHeadVtxBuffer.AddArrowHeadTriangle(edVtPos[0],edVtPos[1],0.2);
			}
			else
			{
				YsVec3 clippedEdVtPos[2];
				if(YSOK==YsClipLineSeg3(
				    clippedEdVtPos[0],clippedEdVtPos[1],edVtPos[0],edVtPos[1],viewPort[0],viewPort[1]))
				{
					selectedConstEdgeVtxBuffer.AddVertex(clippedEdVtPos[0]);
					selectedConstEdgeVtxBuffer.AddVertex(clippedEdVtPos[1]);

					selectedConstEdgeIdxBuffer.Append(shl.GetSearchKey(ceHdArray[idx]));

					selectedConstEdgeArrowHeadVtxBuffer.AddArrowHeadTriangle(clippedEdVtPos[0],clippedEdVtPos[1],0.2);
				}
			}
		}
	}
}

void YsShellExtDrawingBuffer::MakeVertexNormalCache(YsHashTable <YsVec3> &vtxNomHash,const YsShellExt &shl)
{
	for(YsShellPolygonHandle plHd=NULL; YSOK==shl.MoveToNextPolygon(plHd); )
	{
		int nPlVt;
		const YsShellVertexHandle *plVtHd;
		shl.GetVertexListOfPolygon(nPlVt,plVtHd,plHd);

		YsVec3 nom;
		shl.GetNormalOfPolygon(nom,plHd);

		for(int idx=0; idx<nPlVt; ++idx)
		{
			const YsShellExt::VertexAttrib *vtAttr=shl.GetVertexAttrib(plVtHd[idx]);
			if(YSTRUE==vtAttr->IsRound())
			{
				YsVec3 nomSum;
				if(YSOK!=vtxNomHash.FindElement(nomSum,shl.GetSearchKey(plVtHd[idx])))
				{
					nomSum=YsOrigin();
				}
				nomSum+=nom;
				vtxNomHash.UpdateElement(shl.GetSearchKey(plVtHd[idx]),nomSum);
			}
		}
	}
	for(YsShellVertexHandle vtHd=NULL; YSOK==shl.MoveToNextVertex(vtHd); )
	{
		YsVec3 nomSum;
		if(YSOK==vtxNomHash.FindElement(nomSum,shl.GetSearchKey(vtHd)))
		{
			nomSum.Normalize();
			vtxNomHash.UpdateElement(shl.GetSearchKey(vtHd),nomSum);
		}
	}
}

////////////////////////////////////////////////////////////

class YsShellExtDrawingBuffer::RemakeSelectedFaceGroupBufferTask : public YsTask
{
public:
	YsGLVertexBuffer selectedFaceGroupVtxBuffer;
	YsArray <int> selectedFaceGroupIdxBuffer;

	YsArray <YsShellPolygonHandle> plHdArray;
	YsArray <int> indexArray;

	const YsShellExt *shlExtPtr;
	YSBOOL viewPortEnabled;
	YsVec3 viewPort[2];

	virtual void Run(void);  // Implementation of the thread task.
	void AddTriangle(const YsVec3 tri[3],int fgKey);
};

void YsShellExtDrawingBuffer::RemakeSelectedFaceGroupBufferTask::Run(void)
{
	for(YSSIZE_T idx=0; idx<plHdArray.GetN(); ++idx)
	{
		YsArray <YsVec3,16> plVtPos;
		shlExtPtr->GetVertexListOfPolygon(plVtPos,plHdArray[idx]);

		if(3==plVtPos.GetN())
		{
			YsVec3 triVtPos[3]={plVtPos[0],plVtPos[1],plVtPos[2]};
			AddTriangle(triVtPos,indexArray[idx]);
		}
		else if(YSTRUE==YsCheckConvex3(plVtPos.GetN(),plVtPos))
		{
			for(int i=1; i<plVtPos.GetN()-1; i++)
			{
				YsVec3 triVtPos[3]={plVtPos[0],plVtPos[i],plVtPos[i+1]};
				AddTriangle(triVtPos,indexArray[idx]);
			}
		}
		else
		{
			YsSword swd;

			swd.SetInitialPolygon(plVtPos.GetN(),plVtPos,NULL);
			swd.Triangulate();
			for(int i=0; i<swd.GetNumPolygon(); i++)
			{
				const YsArray <YsVec3> *tri=swd.GetPolygon(i);
				if(tri->GetN()>=3)
				{
					const YsVec3 triVtPos[3]={(*tri)[0],(*tri)[1],(*tri)[2]};
					AddTriangle(triVtPos,indexArray[idx]);
				}
			}
		}
	}
}

void YsShellExtDrawingBuffer::RemakeSelectedFaceGroupBufferTask::AddTriangle(const YsVec3 tri[3],int fgKey)
{
	if(YSTRUE!=viewPortEnabled)
	{
		selectedFaceGroupVtxBuffer.AddVertex(tri[0]);
		selectedFaceGroupVtxBuffer.AddVertex(tri[1]);
		selectedFaceGroupVtxBuffer.AddVertex(tri[2]);
		selectedFaceGroupIdxBuffer.Append(fgKey);
	}
	else
	{
		int newNp;
		YsVec3 newp[32]; // 15 should be good enough.  Just in case.
		if(YSOK==YsClipPolygon3(newNp,newp,32,3,tri,viewPort[0],viewPort[1]) && 3<=newNp)
		{
			for(YSSIZE_T idx=1; idx<newNp-1; ++idx)
			{
				YsVec3 tri[3]={newp[0],newp[idx],newp[idx+1]};
				selectedFaceGroupVtxBuffer.AddVertex(tri[0]);
				selectedFaceGroupVtxBuffer.AddVertex(tri[1]);
				selectedFaceGroupVtxBuffer.AddVertex(tri[2]);
				selectedFaceGroupIdxBuffer.Append(fgKey);
			}
		}
	}
}

void YsShellExtDrawingBuffer::RemakeSelectedFaceGroupBuffer(const class YsShellExt &shl,YSSIZE_T nFg,const YsShellExt::FaceGroupHandle fgHdArray[])
{
	selectedFaceGroupVtxBuffer.CleanUp();
	selectedFaceGroupIdxBuffer.CleanUp();

	int nThread=(int)std::thread::hardware_concurrency();
	if(0==nThread)
	{
		nThread=4;
	}
	YsArray <YsShellExtDrawingBuffer::RemakeSelectedFaceGroupBufferTask,32> taskArray(nThread,NULL);
	YsArray <YsTask *,32> taskPtrArray(nThread,NULL);

	for(YSSIZE_T taskIdx=0; taskIdx<nThread; ++taskIdx)
	{
		taskArray[taskIdx].shlExtPtr=&shl;
		taskArray[taskIdx].viewPortEnabled=viewPortEnabled;
		taskArray[taskIdx].viewPort[0]=viewPort[0];
		taskArray[taskIdx].viewPort[1]=viewPort[1];
		taskPtrArray[taskIdx]=&taskArray[taskIdx];
	}

	int taskIdx=0;
	for(YSSIZE_T fgIdx=0; fgIdx<nFg; ++fgIdx)
	{
		YsArray <YsShellPolygonHandle,16> fgPlHd;
		shl.GetFaceGroup(fgPlHd,fgHdArray[fgIdx]);

		const int fgKey=shl.GetSearchKey(fgHdArray[fgIdx]);

		for(YSSIZE_T plIdx=0; plIdx<fgPlHd.GetN(); ++plIdx)
		{
			taskArray[taskIdx].plHdArray.Append(fgPlHd[plIdx]);
			taskArray[taskIdx].indexArray.Append(fgKey);
			++taskIdx;
			if(nThread<=taskIdx)
			{
				taskIdx=0;
			}
		}
	}

#ifdef YS_RUN_PARALLEL
	YsTask::RunParallel(taskPtrArray);
#else
	for(int i=0; i<nThread; ++i) // Will launch in parallel soon.
	{
		taskArray[i].StartLocal();
	}
#endif

	for(YSSIZE_T taskIdx=0; taskIdx<nThread; ++taskIdx)
	{
		selectedFaceGroupVtxBuffer.MoveFrom(taskArray[taskIdx].selectedFaceGroupVtxBuffer);
		selectedFaceGroupIdxBuffer.Append(taskArray[taskIdx].selectedFaceGroupIdxBuffer);
	}
}

////////////////////////////////////////////////////////////

YsColor YsShellExtDrawingBuffer::ColorGradation(const double t)
{
	if(t<0.7f)
	{
		const double s=t/0.7;
		return YsColor(s,0.0,1.0,1.0);
	}
	else
	{
		const double s=(t-0.7f)/0.3f;
		return YsColor(1.0,0.0f,1.0-s,1.0);
	}

}

YsColor YsShellExtDrawingBuffer::ColorGradation(int idx,int n)
{
	if(1<n)
	{
		const double t=(double)idx/(double)(n-1);
		return ColorGradation(t);
	}
	else
	{
		return YsBlue();
	}
}
