/* ////////////////////////////////////////////////////////////

  File Name: ysshellextsearch.cpp
  Copyright (c) 2014 Soji Yamakawa.  All rights reserved.
  http://www.ysflight.com

//////////////////////////////////////////////////////////// */

#include "ysshellextsearch.h"


YsShellExt::SearchTable::SearchTable()
{
	shl=NULL;
	CleanUp();
}

YsShellExt::SearchTable::~SearchTable()
{
}


void YsShellExt::SearchTable::CleanUp(void)
{
	shl=NULL;
	vtKeyToCeHd.CleanUp();
	plKeyToFgHd.CleanUp();
}

void YsShellExt::SearchTable::Attach(const YsShellExt *shl)
{
	CleanUp();
	this->shl=shl;

	YsShellExt::ConstEdgeHandle ceHd=NULL;
	while(YSOK==shl->MoveToNextConstEdge(ceHd))
	{
		ConstEdgeAdded(ceHd);
	}

	YsShellExt::FaceGroupHandle fgHd=NULL;
	while(YSOK==shl->MoveToNextFaceGroup(fgHd))
	{
		FaceGroupAdded(fgHd);
	}

	YsShellExt::VolumeHandle vlHd=NULL;
	while(YSOK==shl->MoveToNextVolume(vlHd))
	{
		VolumeAdded(vlHd);
	}
}

void YsShellExt::SearchTable::Detach(void)
{
	CleanUp();
	this->shl=NULL;
}

void YsShellExt::SearchTable::ConstEdgeAdded(YsShellExt::ConstEdgeHandle ceHd)
{
	if(NULL!=this && NULL!=shl)
	{
		YSSIZE_T nCeVt;
		const YsShellVertexHandle *ceVtHd;
		YSBOOL isLoop;
		if(YSOK==shl->GetConstEdge(nCeVt,ceVtHd,isLoop,ceHd))
		{
			ConstEdgeVertexAdded(ceHd,nCeVt,ceVtHd);
		}
	}
}

void YsShellExt::SearchTable::ConstEdgeVertexAdded(YsShellExt::ConstEdgeHandle ceHd,const YSSIZE_T nCeVt,const YsShellVertexHandle ceVtHd[])
{
	if(NULL!=this && NULL!=shl)
	{
		for(YSSIZE_T idx=0; idx<nCeVt; ++idx)
		{
			unsigned int key[1]=
			{
				shl->GetSearchKey(ceVtHd[idx])
			};
			vtKeyToCeHd.AddElement(1,key,ceHd);
		}
	}
}

void YsShellExt::SearchTable::ConstEdgeAboutToBeDeleted(YsShellExt::ConstEdgeHandle ceHd)
{
	if(NULL!=this && NULL!=shl)
	{
		YSSIZE_T nCeVt;
		const YsShellVertexHandle *ceVtHd;
		YSBOOL isLoop;
		if(YSOK==shl->GetConstEdge(nCeVt,ceVtHd,isLoop,ceHd))
		{
			for(YSSIZE_T idx=0; idx<nCeVt; ++idx)
			{
				unsigned int key[1]=
				{
					shl->GetSearchKey(ceVtHd[idx])
				};
				vtKeyToCeHd.DeleteElement(1,key,ceHd);
			}
		}
	}
}


void YsShellExt::SearchTable::FaceGroupAdded(YsShellExt::FaceGroupHandle fgHd)
{
	if(NULL!=this && NULL!=shl)
	{
		YSSIZE_T nFgPl;
		const YsShellPolygonHandle *fgPlHd;
		if(YSOK==shl->GetFaceGroup(nFgPl,fgPlHd,fgHd))
		{
			for(YSSIZE_T idx=0; idx<nFgPl; ++idx)
			{
				plKeyToFgHd.UpdateElement(shl->GetSearchKey(fgPlHd[idx]),fgHd);
			}
		}
	}
}

void YsShellExt::SearchTable::FaceGroupAboutToBeDeleted(YsShellExt::FaceGroupHandle fgHd)
{
	if(NULL!=this && NULL!=shl)
	{
		YSSIZE_T nFgPl;
		const YsShellPolygonHandle *fgPlHd;
		if(YSOK==shl->GetFaceGroup(nFgPl,fgPlHd,fgHd))
		{
			for(YSSIZE_T idx=0; idx<nFgPl; ++idx)
			{
				plKeyToFgHd.DeleteElement(shl->GetSearchKey(fgPlHd[idx]),fgHd);
			}
		}
	}
}

void YsShellExt::SearchTable::PolygonAddedToFaceGroup(YsShellExt::FaceGroupHandle fgHd,YsShellPolygonHandle plHd)
{
	if(NULL!=this && NULL!=shl)
	{
		plKeyToFgHd.UpdateElement(shl->GetSearchKey(plHd),fgHd);
	}
}

void YsShellExt::SearchTable::PolygonRemovedFromFaceGroup(YsShellExt::FaceGroupHandle fgHd,YsShellPolygonHandle plHd)
{
	if(NULL!=this && NULL!=shl)
	{
		plKeyToFgHd.DeleteElement(shl->GetSearchKey(plHd),fgHd);
	}
}

void YsShellExt::SearchTable::MultiplePolygonAddedToFaceGroup(YsShellExt::FaceGroupHandle fgHd,YSSIZE_T nPl,const YsShellPolygonHandle plHd[])
{
	if(NULL!=this && NULL!=shl)
	{
		for(YSSIZE_T idx=0; idx<nPl; ++idx)
		{
			plKeyToFgHd.UpdateElement(shl->GetSearchKey(plHd[idx]),fgHd);
		}
	}
}

void YsShellExt::SearchTable::MultiplePolygonRemovedFromFaceGroup(YsShellExt::FaceGroupHandle fgHd,YSSIZE_T nPl,YsShellPolygonHandle plHd[])
{
	if(NULL!=this && NULL!=shl)
	{
		for(YSSIZE_T idx=0; idx<nPl; ++idx)
		{
			plKeyToFgHd.DeleteElement(shl->GetSearchKey(plHd[idx]),fgHd);
		}
	}
}


void YsShellExt::SearchTable::VolumeAdded(YsShellExt::VolumeHandle vlHd)
{
	if(NULL!=this && NULL!=shl)
	{
		YSSIZE_T nVlFg;
		const YsShellExt::FaceGroupHandle *vlFgHd;
		if(YSOK==shl->GetVolume(nVlFg,vlFgHd,vlHd))
		{
			for(YSSIZE_T idx=0; idx<nVlFg; ++idx)
			{
				unsigned int key[1]=
				{
					shl->GetSearchKey(vlFgHd[idx])
				};
				fgKeyToVlHd.AddElement(1,key,vlHd);
			}
		}
	}
}

void YsShellExt::SearchTable::VolumeAboutToBeDeleted(YsShellExt::VolumeHandle vlHd)
{
	if(NULL!=this && NULL!=shl)
	{
		YSSIZE_T nVlFg;
		const YsShellExt::FaceGroupHandle *vlFgHd;
		if(YSOK==shl->GetVolume(nVlFg,vlFgHd,vlHd))
		{
			for(YSSIZE_T idx=0; idx<nVlFg; ++idx)
			{
				unsigned int key[1]=
				{
					shl->GetSearchKey(vlFgHd[idx])
				};
				fgKeyToVlHd.DeleteElement(1,key,vlHd);
			}
		}
	}
}

void YsShellExt::SearchTable::MultipleFaceGroupAddedToVolume(YsShellExt::VolumeHandle vlHd,YSSIZE_T nFg,const YsShellExt::FaceGroupHandle fgHd[])
{
	if(NULL!=this && NULL!=shl)
	{
		for(YSSIZE_T fgIdx=0; fgIdx<nFg; ++fgIdx)
		{
			unsigned int key[1]=
			{
				shl->GetSearchKey(fgHd[fgIdx])
			};
			fgKeyToVlHd.AddElement(1,key,vlHd);
		}
	}
}

YSRESULT YsShellExt::SearchTable::FindConstEdgeFromVertex(YSSIZE_T &nCe,const YsShellExt::ConstEdgeHandle *&ceHdPtr,YsShellVertexHandle vtHd) const
{
	if(NULL!=this && NULL!=shl)
	{
		unsigned int n;
		unsigned int key[1]={shl->GetSearchKey(vtHd)};
		if(YSOK==vtKeyToCeHd.FindElement(n,ceHdPtr,1,key))
		{
			nCe=n;
			return YSOK;
		}
	}
	nCe=0;
	ceHdPtr=NULL;
	return YSERR;
}

YsShellExt::FaceGroupHandle YsShellExt::SearchTable::FindFaceGroupFromPolygon(YsShellPolygonHandle plHd) const
{
	if(NULL!=this && NULL!=shl)
	{
		YsShellExt::FaceGroupHandle fgHd;
		if(YSOK==plKeyToFgHd.FindElement(fgHd,shl->GetSearchKey(plHd)))
		{
			return fgHd;
		}
	}
	return NULL;
}

YSRESULT YsShellExt::SearchTable::FindVolumeFromFaceGroup(YSSIZE_T &nVl,const YsShellExt::VolumeHandle *&fgVlHd,YsShellExt::FaceGroupHandle fgHd) const
{
	if(NULL!=this && NULL!=shl)
	{
		unsigned int key[1]=
		{
			shl->GetSearchKey(fgHd)
		};
		return fgKeyToVlHd.FindElement(nVl,fgVlHd,1,key);
	}
	nVl=0;
	return YSERR;
}

