#include "ysfileinfo.h"
#include "ysencoding.h"

YSBOOL YsFileList::MatchPrefix(const YsWString &tst,const YsWString &uPrefix)
{
	const wchar_t asterisk[]={L"*"}; // Workaround for GCC MacOSX specific 64-bit bug
	const wchar_t question[]={L"?"}; // Workaround for GCC MacOSX specific 64-bit bug
	if(0==uPrefix.Strlen() ||
	   0==YsWString::Strcmp(uPrefix,asterisk) ||
	   0==YsWString::Strcmp(uPrefix,question) ||
	   0==YsWString::Strncmp(tst,uPrefix,uPrefix.Strlen()))
	{
		return YSTRUE;
	}
	else
	{
		return YSFALSE;
	}
}

YSBOOL YsFileList::MatchPostfix(const YsWString &tst,const YsWString &uPostfix)
{
	const wchar_t asterisk[]={L"*"}; // Workaround for GCC MacOSX specific 64-bit bug
	const wchar_t question[]={L"?"}; // Workaround for GCC MacOSX specific 64-bit bug
	if(0==uPostfix.Strlen() ||
	   0==YsWString::Strcmp(uPostfix,asterisk) ||
	   0==YsWString::Strcmp(uPostfix,question) ||
	   (tst.Strlen()>=uPostfix.Strlen() && 0==YsWString::Strcmp(tst.GetArray()+tst.Strlen()-uPostfix.Strlen(),uPostfix)))
	{
		return YSTRUE;
	}
	else
	{
		return YSFALSE;
	}
}

YsFileList::YsFileList()
{
}

YsFileList::~YsFileList()
{
}

void YsFileList::Initialize(void)
{
	fileInfo.Clear();
}

void YsFileList::CleanUp(void)
{
	Initialize();
}

YSRESULT YsFileList::TestAndAddFileInfo(const wchar_t fn[],YSBOOL isDir,unsigned long szHigh,unsigned long szLow)
{
	YsWString tst(fn);

	tst.Uncapitalize();
	if(YSTRUE==isDir ||
	   (YSTRUE==MatchPrefix(tst,uPrefix) &&
	    YSTRUE==MatchPostfix(tst,uPostfix)))
	{
		fileInfo.Increment();
		fileInfo.GetEnd().fn.Set(fn);
		YsUnicodeToSystemEncoding(fileInfo.GetEnd().sysFn,fn);
		fileInfo.GetEnd().isDirectory=isDir;
		fileInfo.GetEnd().sizeHigh=szHigh;
		fileInfo.GetEnd().sizeLow=szLow;
	}

	return YSOK;
}

int YsFileList::GetN(void) const
{
	return fileInfo.GetN();
}

YSBOOL YsFileList::IsDirectory(int fileNum) const
{
	if(YSTRUE==fileInfo.IsInRange(fileNum))
	{
		return fileInfo[fileNum].isDirectory;
	}
	return YSFALSE;
}

const char *YsFileList::GetSystemEncodingName(int fileNum) const
{
	if(YSTRUE==fileInfo.IsInRange(fileNum))
	{
		return fileInfo[fileNum].sysFn;
	}
	else
	{
		return NULL;
	}
}

const char *YsFileList::GetSystemEncodingNameWithPath(YsString &nameWithPath,int fileNum) const
{
	if(YSTRUE==fileInfo.IsInRange(fileNum))
	{
		YsWString ful;
		ful.MakeFullPathName(path,fileInfo[fileNum].fn);
		YsUnicodeToSystemEncoding(nameWithPath,ful);
		return nameWithPath;
	}
	else
	{
		nameWithPath.Set("");
		return NULL;
	}
}

const wchar_t *YsFileList::GetFileName(int fileNum) const
{
	if(YSTRUE==fileInfo.IsInRange(fileNum))
	{
		return fileInfo[fileNum].fn;
	}
	else
	{
		return NULL;
	}
}

const wchar_t *YsFileList::GetFileName(YsWString &nameWithPath,int fileNum) const
{
	if(YSTRUE==fileInfo.IsInRange(fileNum))
	{
		nameWithPath.MakeFullPathName(path,fileInfo[fileNum].fn);
		return nameWithPath;
	}
	else
	{
		nameWithPath.Set(NULL);
		return NULL;
	}
}

unsigned long int YsFileList::GetSizeLow(int fileNum) const
{
	if(YSTRUE==fileInfo.IsInRange(fileNum))
	{
		return fileInfo[fileNum].sizeLow;
	}
	return 0;
}

unsigned long int YsFileList::GetSizeHigh(int fileNum) const
{
	if(YSTRUE==fileInfo.IsInRange(fileNum))
	{
		return fileInfo[fileNum].sizeHigh;
	}
	return 0;
}

